/*----------------------------------------------------------------------
  File    : coconad.c
  Contents: COntinuous time ClOsed Neuron Assembly Detection
  Author  : Christian Borgelt
  History : 2013.02.04 file created
            2013.02.19 perfect extension pruning added
            2013.02.20 first version with event filtering completed
            2013.02.21 bug in memory allocation in recurse() fixed
            2013.02.22 some definitions moved to new header coconad.h
            2013.03.07 direction parameter added to sorting functions
            2013.03.19 closed/maximal filtering with repository added
            2013.03.28 adapted to introduction of types ITEM and SUPP
            2013.10.15 checks of return code of isr_report() added
            2013.10.18 optional pattern spectrum collection added
            2013.11.12 item selection file changed to option -R#
            2013.12.03 input format with one train per record added
            2014.02.28 successor buffer eliminated from function sift()
            2014.03.04 adapted to changed data type of train length
            2014.05.12 option -F# added (support border for filtering)
            2014.07.11 option -q added (item sequences instead of sets)
            2014.07.15 sequence mining with event and train filtering
            2014.07.17 sequence mining with repository filtering added
            2014.07.22 bug in perfect exts. with maximal item sets fixed
            2014.08.26 adapted to modified item set reporter interface
            2014.08.28 functions ccn_data() and ccn_report() added
            2014.10.24 changed from LGPL license to MIT license
            2014.10.25 influence region overlap support as alternative
            2014.10.30 overlap support improved (INFINITY -> DBL_MAX)
            2014.12.16 interval list intersection variant added
            2014.12.17 reporting in repository recursion simplified
            2015.09.01 computation of memory size for recursion fixed
            2016.03.22 combined integer and graded support added
            2016.03.23 bug in function isect() fixed (empty intersect.)
            2016.03.26 bug in mixed support fixed (no list intersection)
            2016.09.10 changed to strict time order for sequences
            2016.10.02 extended to sequences with item repetitions
            2016.10.03 closed/maximal w.r.t. appending items added
            2016.10.11 bug in sequence mining fixed (memory sizes)
            2016.10.12 permutation and sequence mining much improved
            2016.10.15 non-overlapping minimal windows support added
            2016.10.18 repository filtering for general sequences added
            2016.10.21 closed/maximal filtering with array repo. added
            2016.10.23 moved repository selection to operation mode
            2016.11.16 coconad miner object and interface introduced
----------------------------------------------------------------------*/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <stdarg.h>
#include <limits.h>
#include <float.h>
#include <time.h>
#include <assert.h>
#ifdef CCN_OVERLAP
#define SUPP double
#endif
#ifndef ISR_PATSPEC
#define ISR_PATSPEC
#endif
#ifndef ISR_CLOMAX
#define ISR_CLOMAX
#endif
#ifdef CCN_MAIN
#ifndef TRN_READ
#define TRN_READ
#endif
#ifndef TA_READ
#define TA_READ
#endif
#endif
#include "train.h"
#ifdef CCN_MAIN
#ifndef PSP_REPORT
#define PSP_REPORT
#endif
#endif
#ifdef CCN_ABORT
#include "sigint.h"
#endif
#include "patspec.h"
#include "coconad.h"
#ifndef CCN_OVERLAP
#include "cm4seqs.h"
#endif
#include "cmfilter.h"
#ifdef CCN_MAIN
#include "params.h"
#include "error.h"
#endif
#ifdef STORAGE
#include "storage.h"
#endif

/*----------------------------------------------------------------------
  Preprocessor Definitions
----------------------------------------------------------------------*/
#define PRGNAME     "coconad"
#ifdef CCN_OVERLAP              /* if influence map overlap support */
#define DESCRIPTION "COntinuous time ClOsed Neuron Assembly Detection" \
                    " (overlap support)"
#else                           /* if standard (MIS) support type */
#define DESCRIPTION "COntinuous time ClOsed Neuron Assembly Detection"
#endif
#define VERSION     "version 2.33 (2016.11.21)        " \
                    "(c) 2013-2016   Christian Borgelt"

/* --- error codes --- */
/* error codes   0 to  -4 defined in tract.h */
#define E_STDIN      (-5)       /* double assignment of stdin */
#define E_OPTION     (-6)       /* unknown option */
#define E_OPTARG     (-7)       /* missing option argument */
#define E_ARGCNT     (-8)       /* too few/many arguments */
#define E_TARGET     (-9)       /* invalid target type */
#define E_SIZE      (-10)       /* invalid item set size */
#define E_SUPPORT   (-11)       /* invalid item set support */
#define E_WIDTH     (-12)       /* invalid window width/max. distance */
#define E_VARIANT   (-13)       /* invalid algorithm variant */
#define E_SIMILTY   (-14)       /* invalid similarity measure */
/* error codes -15 to -25 defined in train.h */

#define HEAPMIN      15         /* minimum number of trains for heap */

#ifndef QUIET                   /* if not quiet version, */
#define MSG         fprintf     /* print messages */
#define XMSG        if (ccn->mode & CCN_VERBOSE) fprintf
#define CLOCK(t)    ((t) = clock())
#else                           /* if quiet version, */
#define MSG(...)    ((void)0)   /* suppress messages */
#define XMSG(...)   ((void)0)
#define CLOCK(t)    ((void)0)
#endif

#define SEC_SINCE(t)  ((double)(clock()-(t)) /(double)CLOCKS_PER_SEC)

/*----------------------------------------------------------------------
  Type Definitions
----------------------------------------------------------------------*/
#ifdef CCN_OVERLAP              /* if influence map overlap support */

typedef struct {                /* --- interval list --- */
  ITEM         item;            /* item identifier (last item in set) */
  SUPP         supp;            /* support of the item (or item set) */
  IVAL         ivals[1];        /* array of intervals */
} IVLIST;                       /* (interval list) */

typedef struct {                /* --- extended interval list --- */
  ITEM         item;            /* item identifier (last item in set) */
  SUPP         all;             /* support (cover   size, all items) */
  SUPP         one;             /* extent  (carrier size, one item) */
  size_t       off;             /* offset to second TID list */
  double       sim;             /* item cover similarity measure */
  IVAL         ivals[2];        /* array of intervals */
} XIVLIST;                      /* (extended interval list) */

typedef double SIMFN (SUPP s, SUPP r, SUPP n, double *p);

#endif  /* #ifdef CCN_OVERLAP */

typedef struct {                /* --- train data for recursion --- */
  TRAIN        **trains;        /* trains of items in set */
  TRAIN        **exts;          /* trains of extension items */
  TRAIN        **excl;          /* trains of excluded items */
  void         *mem;            /* train memory (for filtering) */
  size_t       size;            /* total size of train memory */
  IVAL         ivals[1];        /* intervals for event filtering */
} TRAINDB;                      /* (train database) */

struct _recdata;                /* pattern reporting function */
typedef int REPORTFN (COCONAD *ccn, SUPP smax);

struct _coconad {               /* --- coconad miner --- */
  int          target;          /* target type (e.g. closed/maximal) */
  double       width;           /* width of the influence region */
  SUPP         supp;            /* absolute minimum support */
  ITEM         zmin;            /* minimum size of a rule/item set */
  ITEM         zmax;            /* maximum size of a rule/item set */
  ITEM         xmax;            /* maximum size for extensions */
  int          algo;            /* variant of coconad algorithm */
  int          mode;            /* operation mode (e.g. pruning) */
  TRAINSET     *trnset;         /* train set */
  ISREPORT     *report;         /* item set reporter */
  SUPP         limit;           /* lower bound for support (pexcube) */
  TID          max;             /* maximum size/length of a train */
  ITEM         hsize;           /* current heap size (for sequences) */
  const double **heap;          /* heap for support computation */
  TRAIN        **pexs;          /* trains of perfect extensions */
  TRAIN        **excl;          /* trains of excluded items */
  ITEM         *used;           /* indices of first item/train use */
  REPORTFN     *repfn;          /* pattern reporting function */
  CMFILTER     *cmf;            /* closed/maximal sequence filter */
  #ifndef CCN_OVERLAP           /* if standard (MIS) support */
  CM4SEQS      *cmx;            /* closed/maximal sequence tree chain */
  #else                         /* if influence map overlap support */
  SUPP         total;           /* maximal theoret. possible support */
  SIMFN        *simfn;          /* similarity measure function */
  double       simps[8];        /* similarity measure parameters */
  double       sim;             /* minimum similarity measure value */
  double       mix;             /* mixing coefficient for standard */
  #endif                        /* and influence map overlap support */
};                              /* (coconad miner) */

/*----------------------------------------------------------------------
  Constants
----------------------------------------------------------------------*/
#if !defined QUIET && defined CCN_MAIN
/* --- error messages --- */
static const char *errmsgs[] = {
  /* E_NONE      0 */  "no error",
  /* E_NOMEM    -1 */  "not enough memory",
  /* E_FOPEN    -2 */  "cannot open file %s",
  /* E_FREAD    -3 */  "read error on file %s",
  /* E_FWRITE   -4 */  "write error on file %s",
  /* E_STDIN    -5 */  "double assignment of standard input",
  /* E_OPTION   -6 */  "unknown option -%c",
  /* E_OPTARG   -7 */  "missing option argument",
  /* E_ARGCNT   -8 */  "wrong number of arguments",
  /* E_TARGET   -9 */  "invalid target pattern type '%c'",
  /* E_SIZE    -10 */  "invalid item set size %"ITEM_FMT,
  /* E_SUPPORT -11 */  "invalid minimum support %"SUPP_FMT,
  /* E_WIDTH   -12 */  "invalid window width/maximum distance %g",
  /* E_VARIANT -13 */  "invalid coconad variant '%c'",
  #ifdef CCN_OVERLAP
  /* E_SIMILTY -14 */  "invalid similarity measure '%c'",
  #else
  /*           -14 */  NULL,
  #endif
  /* E_NOITEMS -15 */  "no (frequent) items found",
  /*           -16 */  "unknown error"
};
#endif

/*----------------------------------------------------------------------
  Global Variables
----------------------------------------------------------------------*/
#ifdef CCN_MAIN                 /* if stand-alone program */
#ifndef QUIET
static CCHAR    *prgname;       /* program name for error messages */
#endif
static TABREAD  *tread  = NULL; /* table/train reader */
static ITEMBASE *ibase  = NULL; /* item base */
static TRAINSET *trnset = NULL; /* train set */
static ISREPORT *report = NULL; /* item set reporter */
static TABWRITE *twrite = NULL; /* table writer for pattern spectrum */
static RSUPP    *border = NULL; /* support border for filtering */
static COCONAD *coconad = NULL; /* coconad miner object */
#endif

/*----------------------------------------------------------------------
  Similarity Measure Functions
----------------------------------------------------------------------*/
#ifdef CCN_OVERLAP              /* if influence map overlap support */

static double constant        (SUPP s, SUPP r, SUPP n, double *p)
{ return 0; }

static double russel_rao      (SUPP s, SUPP r, SUPP n, double *p)
{ return (double)s/(double)n; }
/* S_R = s/n */

static double kulczynski      (SUPP s, SUPP r, SUPP n, double *p)
{ SUPP q = r-s; return (q > 0) ? (double)s/(double)q : INFINITY; }
/* S_K = s/q = s/(r-s) */

static double jaccard         (SUPP s, SUPP r, SUPP n, double *p)
{ return (double)s/(double)r; }
/* S_J = s/r */

static double dice            (SUPP s, SUPP r, SUPP n, double *p)
{ return (double)(s+s)/(double)(r+s); }
/* S_D = 2s/(r+s) */

static double sokal_sneath_1  (SUPP s, SUPP r, SUPP n, double *p)
{ return (double)s/(double)(r+r-s); }
/* S_S = s/(r+q) = s/(2r-s) */

static double sokal_michener  (SUPP s, SUPP r, SUPP n, double *p)
{ return (double)(s+(n-r))/(double)n; }
/* S_M = (s+z)/n */

static double faith           (SUPP s, SUPP r, SUPP n, double *p)
{ return ((double)s+0.5*(double)(n-r))/(double)n; }
/* S_F = (s+z/2)/n */

static double rogers_tanimoto (SUPP s, SUPP r, SUPP n, double *p)
{ SUPP q = r-s; return (double)(n-q)/(double)(n+q); }
/* S_T = (s+z)/(n+q) = (n-q)/(n+q) */

static double sokal_sneath_2  (SUPP s, SUPP r, SUPP n, double *p)
{ SUPP sz = s+(n-r); return (double)(sz+sz)/(double)(n+sz); }
/* S_N = 2(s+z)/(n+s+z) */

static double sokal_sneath_3  (SUPP s, SUPP r, SUPP n, double *p)
{ SUPP q = r-s; return (q > 0) ? (double)(n-q)/(double)q : INFINITY; }
/* S_O = (s+z)/q = (m-q)/q */

static double baroni_buser    (SUPP s, SUPP r, SUPP n, double *p)
{ double x = sqrt((double)s*(n-r)); return (x+(double)s)/(x+(double)r);}
/* S_B = (x+s)/(x+r) */

/*--------------------------------------------------------------------*/

static double generic_0       (SUPP s, SUPP r, SUPP n, double *p)
{                               /* --- generic similarity measure */
  double x = sqrt((double)s*(double)r);
  double d = p[4]*(double)s +p[5]*(double)r +p[6]*(double)n +p[7]*x;
  if (fabs(d) <= DBL_MIN) return INFINITY;
  return (p[0]*s +p[1]*r +p[2]*n +p[3]*x) /d;
}  /* generic_0() */
/* S = (c_0s +c_1z +c_2n +c_3x) / (c_4s +c_5z +c_6n +c_7x) */

/*--------------------------------------------------------------------*/

static double generic_1       (SUPP s, SUPP r, SUPP n, double *p)
{                               /* --- generic similarity measure */
  double d = p[4]*(double)s +p[5]*(double)r +p[6]*(double)n;
  if (fabs(d) <= DBL_MIN) return INFINITY;
  return (p[0]*(double)s +p[1]*(double)r +p[2]*(double)n) /d;
}  /* generic_1() */
/* S = (c_0s +c_1z +c_2n) / (c_4s +c_5z +c_6n) */

/*--------------------------------------------------------------------*/

static double generic_2       (SUPP s, SUPP r, SUPP n, double *p)
{                               /* --- generic similarity measure */
  double d = p[4]*(double)s +p[5]*(double)r +p[6]*(double)n;
  return (fabs(d) > DBL_MIN)
       ? (p[0]*(double)s +p[1]*(double)r) /d : INFINITY;
}  /* generic_2() */
/* S = (c_0s +c_1z) / (c_4s +c_5z +c_6n) */

/*--------------------------------------------------------------------*/

static double generic_3       (SUPP s, SUPP r, SUPP n, double *p)
{                               /* --- generic similarity measure */
  double d = p[4]*(double)s +p[5]*(double)r +p[6]*(double)n;
  return (fabs(d) > DBL_MIN) ? p[0]*(double)s /d : INFINITY;
}  /* generic_3() */
/* S = c_0s / (c_4s +c_5z +c_6n) */

/*--------------------------------------------------------------------*/

static SIMFN *simfns[] = {
  /* CCN_NONE             0 */  constant,
  /* CCN_RUSSEL_RAO       1 */  russel_rao,
  /* CCN_KULCZYNSKI       2 */  kulczynski,
  /* CCN_JACCARD          3 */  jaccard,
  /* CCN_DICE             4 */  dice,
  /* CCN_SOKAL_SNEATH_1   5 */  sokal_sneath_1,
  /* CCN_SOKAL_MICHENER   6 */  sokal_michener,
  /* CCN_FAITH            7 */  faith,
  /* CCN_ROGERS_TANIMOTO  8 */  rogers_tanimoto,
  /* CCN_SOKAL_SNEATH_2   9 */  sokal_sneath_2,
  /* CCN_SOKAL_SNEATH_3  10 */  sokal_sneath_3,
  /* CCN_BARONI_BUSER    11 */  baroni_buser,
  /* CCN_GENERIC         12 */  generic_0,
};

#endif  /* #ifdef CCN_OVERLAP */
/*----------------------------------------------------------------------
  Interval List Functions
----------------------------------------------------------------------*/
#ifdef CCN_OVERLAP              /* if influence map overlap support */

static size_t trn2ivals (TRAIN *train, double width, IVLIST *ivl)
{                               /* --- turn train into list of ivals. */
  const double *p;              /* to traverse the points */
  IVAL         *d;              /* to traverse the intervals */
  double       w;               /* +/- offset for intervals */
  double       x;               /* start of the next interval */
  double       supp;            /* support of the interval list */

  assert(train && ivl && (width > 0));
  ivl->item = trn_item(train);  /* check the function arguments */
  ivl->supp = 0;                /* initialize the interval list */
  d = ivl->ivals;               /* and get the first interval */
  p = trn_points(train);        /* get the points of the train */
  if (*p >= TRN_END) { d->beg = d->end = TRN_END; return 0; }
  supp   = 0;  w = 0.5 *width;  /* compute +/- width for intervals */
  d->beg = *p -w;               /* and with it the boundaries */
  d->end = *p +w;               /* of the first interval */
  while (*++p < TRN_END) {      /* traverse the remaining points */
    x = *p -w;                  /* compute start of new interval */
    if (x > d->end) {           /* if the intervals are disjoint, */
      supp += d->end -d->beg;   /* add interval length to support */
      (++d)->beg = x;           /* and start a new output interval */
    }
    d->end = *p +w;             /* set/update the interval end */
  }
  supp += d->end -d->beg; d++;  /* add last interval to support */
  d->beg = d->end = TRN_END;    /* store a sentinel at the end */
  ivl->supp = (SUPP)(supp/width); /* compute and store support */
  return (size_t)(d-ivl->ivals);/* return the number of intervals */
}  /* trn2ivals() */

/*--------------------------------------------------------------------*/

static size_t trn2xivls (TRAIN *train, double width,
                         double beg, double end, XIVLIST *ivl)
{                               /* --- turn train into list of ivals. */
  const double *p;              /* to traverse the points */
  IVAL         *d;              /* to traverse the intervals */
  double       w;               /* +/- offset for intervals */
  double       x;               /* start of the next interval */
  double       supp;            /* support of the interval list */

  assert(train && ivl && (width > 0));
  ivl->item = trn_item(train);  /* check the function arguments */
  ivl->all  = ivl->one = 0;     /* initialize the interval list */
  d = ivl->ivals;               /* and get the first interval */
  p = trn_points(train);        /* get the points of the train */
  if (*p >= TRN_END) { d->beg = d->end = TRN_END; d++;
                       d->beg = d->end = TRN_END; return 0; }
  supp   = 0;  w = 0.5 *width;  /* compute +/- width for intervals */
  d->beg = *p -w;               /* and with it the boundaries */
  d->end = *p +w;               /* of the first interval */
  if (d->beg < beg) d->beg = beg;
  if (d->end > end) d->end = end;
  while (*++p < TRN_END) {      /* traverse the remaining points */
    x = *p -w;                  /* compute start of new interval */
    if (x < beg) x = beg;       /* and clamp it to allowed region */
    if (x > d->end) {           /* if the intervals are disjoint, */
      supp += d->end -d->beg;   /* add interval length to support */
      (++d)->beg = x;           /* and start a new output interval */
    }
    x = *p +w;                  /* compute end of new interval */
    if (x > end) x = end;       /* and clamp it to allowed region */
    d->end = x;                 /* set/update the interval end */
  }
  supp += d->end -d->beg; d++;  /* add last interval to support */
  d->beg = d->end = TRN_END;    /* store a sentinel at the end */
  ivl->all = (SUPP)(supp/width);/* compute support and offset */
  ivl->one = ivl->all;          /* and offset to second list */
  ivl->off = (size_t)(++d -ivl->ivals);
  memcpy(d, ivl->ivals, ivl->off *sizeof(IVAL));
  return 2*ivl->off-1;          /* create a second copy and */
}  /* trn2xivls() */            /* return the number of intervals */

/*--------------------------------------------------------------------*/
#ifndef NDEBUG

static void ivl_show (IVLIST **ivls, ITEM n)
{                               /* --- show a list of interval lists */
  ITEM i;                       /* loop variable */
  IVAL *p;                      /* to traverse the intervals */

  assert(ivls && (n >= 0));     /* check the function arguments */
  for (i = 0; i < n; i++) {     /* traverse the interval lists */
    printf("%2"ITEM_FMT, ivls[i]->item);
    printf(" %"SUPP_FMT, ivls[i]->supp);
    for (p = ivls[i]->ivals; p->beg < TRN_END; p++) {
      printf(" [%g,%g]", p->beg, p->end); }
    printf("\n");               /* print the item, the support, */
  }                             /* and the list of intervals */
  printf("%d interval list(s)\n", n);
}  /* ivl_show() */

/*--------------------------------------------------------------------*/

static void xivl_show (XIVLIST **ivls, ITEM n)
{                               /* --- show a list of interval lists */
  ITEM i;                       /* loop variable */
  IVAL *p;                      /* to traverse the intervals */

  assert(ivls && (n >= 0));     /* check the function arguments */
  for (i = 0; i < n; i++) {     /* traverse the interval lists */
    printf("%2"ITEM_FMT, ivls[i]->item);
    printf(" %"SUPP_FMT, ivls[i]->all);
    printf(" %"SUPP_FMT, ivls[i]->one);
    printf(" %g\n",      ivls[i]->sim);
    printf("all: ");            /* print item and support values */
    for (p = ivls[i]->ivals; p->beg < TRN_END; p++) {
      printf(" [%g,%g]", p->beg, p->end); }
    printf("\none: ");          /* print the list of intervals */
    for (p++; p->beg < TRN_END; p++) {
      printf(" [%g,%g]", p->beg, p->end); }
    printf("\n");               /* print the list of intervals */
  }                             /* (support and carrier) */
  printf("%d interval list(s)\n", n);
}  /* xivl_show() */

#endif
#endif                          /* #ifdef CCN_OVERLAP .. */
/*----------------------------------------------------------------------
  Support Computation Functions
----------------------------------------------------------------------*/

static void sift (const double *heap[], ITEM lft, ITEM rgt)
{                               /* --- let element sift down in heap */
  ITEM  i;                      /* index of first successor in heap */
  const double *t;              /* buffer for element to sift down */

  assert(heap && (lft >= 0) && (rgt > lft));
  t = heap[lft];                /* note the sift element */
  i = lft +lft +1;              /* compute index of first successor */
  do {                          /* sift loop */
    if ((i          < rgt)      /* if second successor exists and */
    &&  (*heap[i+1] < *heap[i]))/* it is smaller than the first, */
      i++;                      /* go to the second successor */
    if (*heap[i] >= *t)         /* if the successor is no less */
      break;                    /* than the sift element, abort */
    heap[lft] = heap[i];        /* let the successor ascend in heap */
    lft = i;                    /* note index of current element */
    i  += i+1;                  /* compute index of first successor */
  } while (i <= rgt);           /* while still within heap */
  heap[lft] = t;                /* store the sift element */
}  /* sift() */

/*--------------------------------------------------------------------*/
#ifndef CCN_OVERLAP             /* if standard (MIS) support */
static SUPP support (TRAIN **trains, ITEM n, double width,
                     const double *heap[])
#else                           /* if influence map overlap support */
static SUPP intsupp (TRAIN **trains, ITEM n, double width,
                     const double *heap[])
#endif                          /* (for combined int+graded support) */
{                               /* --- compute item set support */
  ITEM   i, k;                  /* loop variables */
  SUPP   supp;                  /* support of the item set */
  double min, max, m;           /* minimum/maximum point in heap */
  int    mis;                   /* flag for maximum independent set */

  assert(trains && (n > 0) && heap); /* check the function arguments */
  if (n <= 1)                   /* if there is only one train, */
    return (SUPP)trn_len(trains[0]);      /* return its length */
  mis   = (width >= 0);         /* get the support type from the sign */
  width = fabs(width);          /* ensure non-negative window width */
  supp  = 0;                    /* initialize the item set support */
  if (n <= 2) {                 /* if there are only two trains */
    heap[0] = trn_points(trains[0]);
    heap[1] = trn_points(trains[1]);
    while ((*heap[0] < TRN_END) /* traverse the two trains */
    &&     (*heap[1] < TRN_END)) {
      double d = *heap[0] -*heap[1];
      if (fabs(d) > width)      /* if the points are not coincident, */
        heap[(d > 0) ? 1 : 0]++;/* skip the earlier point */
      else if (mis) {           /* if maximum independent set support */
        heap[0]++; heap[1]++; supp += 1; }
      else {                    /* if non-overlapping minimal windows */
        k = (d > 0) ? 1 : 0;    /* get train index of smaller time */
        while (*++heap[k] <= *heap[1-k]);
        heap[1-k]++; supp += 1; /* skip until after later time */
      }                         /* (i.e. to end of minimal window) */
    }                           /* and count the point coincidences */
    return supp;                /* return the computed support */
  }
  *heap = trn_points(*trains);  /* copy trains to a heap/array */
  for (max = *heap[i = 0]; ++i < n; ) {
    heap[i] = trn_points(trains[i]);
    if (*heap[i] > max) max = *heap[i];
  }                             /* find maximum value in heap/array */
  if (n < HEAPMIN) {            /* if there are only few trains */
    while (max < TRN_END) {     /* while no train is exhausted */
      for (min = *heap[i = k = 0]; ++i < n; )
        if (*heap[i] < min)     /* find the train */
          min = *heap[k = i];   /* with the minimum point */
      if (max -min > width) {   /* if points are not coincident */
        if (*++heap[k] > max) max = *heap[k];
        continue;               /* skip only the minimum point, */
      }                         /* otherwise (points are coincident) */
      supp += 1;                /* count coincidence for the support */
      if (mis) {                /* if maximum independent set support */
        for (i = 0; i < n; i++) /* skip all points in the array */
          if (*++heap[i] > max) max = *heap[i]; }
      else {                    /* if non-overlapping minimal windows */
        for (m = max, i = 0; i < n; i++) {
          while (*++heap[i] <= m);
          if (*heap[i] > max) max = *heap[i];
        }                       /* skip up to end of minimal window */
      }                         /* and find new maximum time */
    } }
  else {                        /* if there are many items */
    if (max >= TRN_END)         /* if maximum is not finite, */
      return 0;                 /* at least one train is empty */
    for (i = k = n-- >> 1; --i >= 0; )
      sift(heap, i, n);         /* build the point heap */
    while (1) {                 /* while no train is exhausted */
      if (max-*heap[0] > width){/* if points are not coincident */
        if (*++heap[0] > max) max = *heap[0];
        if (max >= TRN_END) break;
        sift(heap, 0, n);       /* skip the minimum point and */
        continue;               /* let the new point sift down */
      }                         /* otherwise (points are coincident) */
      supp += 1;                /* count coincidence for the support */
      if (mis) {                /* if maximum independent set support */
        for (i = 0; i <= n; i++)/* skip all points in the heap */
          if (*++heap[i] > max) max = *heap[i]; }
      else {                    /* if non-overlapping minimal windows */
        for (m = max, i = 0; i <= n; i++) {
          while (*++heap[i] <= m);
          if (*heap[i] > max) max = *heap[i];
        }                       /* skip up to end of minimal window */
      }                         /* and find new maximum time */
      if (max >= TRN_END) break;/* check for an empty train */
      for (i = k; --i >= 0; )   /* rebuild the point heap */
        sift(heap, i, n);       /* (all entries are changed) */
    }
  }
  return supp;                  /* return the computed support */
}  /* support() */

/*--------------------------------------------------------------------*/
#ifndef CCN_OVERLAP             /* if standard (MIS) support */

static SUPP seqsupp (TRAIN **trains, ITEM n, double width,
                     const double *curr[])
{                               /* --- compute item sequence support */
  ITEM   i;                     /* loop variable */
  SUPP   supp;                  /* support of the item sequence */
  double max;                   /* maximum point in sequence */
  int    mis;                   /* flag for maximum independent set */

  assert(trains && (n > 0) && curr); /* check the function arguments */
  if (n <= 1)                   /* if there is only one train, */
    return (SUPP)trn_len(trains[0]);      /* return its length */
  mis   = (width >= 0);         /* get the support type from the sign */
  width = fabs(width);          /* ensure non-negative window width */
  supp  = 0;                    /* initialize the sequence support */
  if (n <= 2) {                 /* if there are only two trains */
    curr[0] = trn_points(trains[0]);
    curr[1] = trn_points(trains[1]);
    while ((*curr[0] < TRN_END) /* traverse the two trains */
    &&     (*curr[1] < TRN_END)) {
      double d = *curr[1] -*curr[0];
      if      (d <= 0)    curr[1]++;
      else if (d > width) curr[0]++;
      else if (mis) {           /* if maximum independent set support */
        curr[0]++; curr[1]++; supp += 1; }
      else {                    /* if non-overlapping minimal windows */
        while (*++curr[0] <= *curr[1]);
        curr[1]++; supp += 1;   /* skip until after later time */
      }                         /* (i.e. to end of minimal window) */
    }                           /* count the sequence occurrences */
    return supp;                /* return the computed support */
  }
  for (i = 0; i < n; i++)       /* traverse and copy the trains */
    curr[i] = trn_points(trains[i]);
  while (*curr[0] < TRN_END) {  /* support computation loop */
    for (i = 1; i < n; i++) {   /* ensure item succession in time */
      while (*curr[i] <= *curr[i-1])
        curr[i]++;              /* skip times not in correct order */
      if (*curr[i] >= TRN_END) { i++; break; }
    }                           /* check for end of train reached */
    max = *curr[i-1];           /* get the end of the sequence */
    if (max >= TRN_END) break;  /* if beyond end of trains, abort */
    if (max -*curr[0] > width)  /* if points are not coincident, */
      curr[0]++;                /* get the next starting point */
    else if (mis) {             /* if maximum independent set support */
      for (i = 0; i < n; i++)   /* skip the points of the instance */
        curr[i]++;              /* (the points cannot be used again) */
      supp += 1; }              /* count the current instance */
    else {                      /* if non-overlapping minimal windows */
      for (i = 0; i < n; i++)   /* skip the points in minimal window */
        while (*++curr[i] <= max);
      supp += 1;                /* count the current instance */
    }
  }
  return supp;                  /* return the computed support */
}  /* seqsupp() */

/*--------------------------------------------------------------------*/
#else                           /* if influence map overlap support */

static SUPP support (TRAIN **trains, ITEM n, double width,
                     const double *heap[])
{                               /* --- compute item set support */
  const double *p;              /* to traverse the points */
  ITEM   i, k;                  /* loop variables */
  double supp;                  /* support of the item set */
  double min, max;              /* minimum/maximum point in heap */
  double a, b, c;               /* start and end of intervals */
  double w;                     /* half the window/region width */

  assert(trains                 /* check the function arguments */
  &&    (n > 0) && (width >= 0) && heap);
  w    = 0.5*width;             /* get half the window width and */
  supp = 0;                     /* initialize the item set support */
  if (n <= 1) {                 /* if there is only one train */
    p = trn_points(trains[0]);  /* and the points of the train */
    if (*p >= TRN_END) return 0;/* if there are no points, abort */
    a = *p -w; b = *p +w;       /* compute start/end of 1st interval */
    while (*++p < TRN_END) {    /* traverse the points of the train */
      c = *p -w;                /* compute start of new interval */
      if (c > b) {              /* if the intervals are disjoint, */
        supp += b-a;            /* sum length of preceding interval */
        a = c;                  /* and start a new interval */
      }
      b = *p +w;                /* update end of current interval */
    }                           /* add length of the last interval */
    return (SUPP)((supp +(b-a)) /width);
  }                             /* return the computed support */
  *heap = trn_points(*trains);  /* copy trains to a heap/array */
  for (max = *heap[i = 0]; ++i < n; ) {
    heap[i] = trn_points(trains[i]);
    if (*heap[i] > max) max = *heap[i];
  }                             /* find maximum value in heap/array */
  if (max >= TRN_END)           /* if maximum is not finite, */
    return 0;                   /* at least one train is empty */
  if (n >= HEAPMIN)             /* if there are many trains */
    for (i = k = n-- >> 1; --i >= 0; )
      sift(heap, i, n);         /* build the point heap */
  a = b = -DBL_MAX;             /* initialize the first interval */
  while (1) {                   /* support computation loop */
    min = *heap[k = 0];         /* get the minimum time (if heap) */
    if (n < HEAPMIN) {          /* if there are only few trains */
      for (i = 0; ++i < n; )    /* traverse the remaining trains */
        if (*heap[i] < min) min = *heap[k = i];
    }                           /* find heap with minimum point */
    if (max-min <= width) {     /* if the times are synchronous */
      c = max -w;               /* compute start of new interval */
      if (c > b) {              /* if the intervals are disjoint */
        supp += b-a;            /* sum length of preceding interval */
        a = c;                  /* and start a new interval */
      }
      b = min +w;               /* update end of current interval */
    }                           /* ([a,b] is always current interval) */
    if (*++heap[k] > max)       /* skip the processed point */
      max = *heap[k];           /* and update the maximum */
    if (max >= TRN_END) break;  /* check for the end of a train */
    if (n < HEAPMIN) sift(heap, 0, n-1);
  }                             /* restore heap structure */
  return (SUPP)((supp +(b-a)) /width);
}  /* support() */              /* return the computed support */

/*--------------------------------------------------------------------*/

static size_t isect (IVLIST *dst, IVLIST *src1, IVLIST *src2,
                     double width)
{                               /* --- intersect two interval lists */
  IVAL   *s1, *s2, *d;          /* to traverse sources and dest. */
  double a, z;                  /* start and end of an interval */
  double supp;                  /* support of the intersection */

  assert(dst && src1 && src2    /* check the function arguments */
  &&    (src1->ivals[0].beg < TRN_END)
  &&    (src2->ivals[0].beg < TRN_END) && (width >= 0));
  dst->item = src1->item;       /* copy the first item */
  s1 = src1->ivals;             /* get the source interval lists */
  s2 = src2->ivals;             /* and the destination interval list */
  d  = dst->ivals; supp = 0.0;  /* and initialize the support */
  while (1) {                   /* trans. id list intersection loop */
    if ((s1->beg < s2->end)     /* if the two current intervals */
    &&  (s1->end > s2->beg)) {  /* overlap (intersection not empty) */
      d->beg = (s1->beg > s2->beg) ? s1->beg : s2->beg;
      d->end = (s1->end < s2->end) ? s1->end : s2->end;
      break;                    /* set the first interval */
    }                           /* and abort the search loop */
    if (s1->end < s2->end) { if ((++s1)->beg >= TRN_END) break; }
    else                   { if ((++s2)->beg >= TRN_END) break; }
  }                             /* skip interval ending earlier */
  if ((s1->beg >= TRN_END) || (s2->beg >= TRN_END)) {
    d->beg = d->end = TRN_END; dst->supp = 0.0; return 0; }
  while (1) {                   /* trans. id list intersection loop */
    if ((s1->beg < s2->end)     /* if the two current intervals */
    &&  (s1->end > s2->beg)) {  /* overlap (intersection not empty) */
      a = (s1->beg > s2->beg) ? s1->beg : s2->beg;
      z = (s1->end < s2->end) ? s1->end : s2->end;
      if (a > d->end) {         /* if the intervals are disjoint */
        supp += d->end -d->beg; /* add interval length to support */
        (++d)->beg = a;         /* and start a new output interval */
      }
      d->end = z;               /* set/update the interval end */
    }
    if (s1->end < s2->end) { if ((++s1)->end >= TRN_END) break; }
    else                   { if ((++s2)->end >= TRN_END) break; }
  }                             /* skip interval ending earlier */
  supp += d->end -d->beg; d++;  /* add last output interval */
  d->beg = d->end = TRN_END;    /* store a sentinel at the list end */
  dst->supp = (SUPP)(supp/width);  /* compute and store the support */
  return (size_t)(d-dst->ivals);/* return the number of intervals */
}  /* isect() */

/*--------------------------------------------------------------------*/

static void xisect (XIVLIST *dst, XIVLIST *src1, XIVLIST *src2,
                    double width)
{                               /* --- intersect two interval lists */
  IVAL   *s1, *s2, *d;          /* to traverse sources and dest. */
  double a, z;                  /* start and end of an interval */
  double supp;                  /* support of the intersection */

  assert(dst && src1 && src2    /* check the function arguments */
  &&    (src1->ivals[0].beg < TRN_END)
  &&    (src2->ivals[0].beg < TRN_END) && (width >= 0));
  dst->item = src1->item;       /* copy the first item */
  s1 = src1->ivals;             /* get sources and destination */
  s2 = src2->ivals;             /* (both refer to intersection lists) */
  d  = dst->ivals; supp = 0.0;  /* and initialize the support */
  while (1) {                   /* trans. id list intersection loop */
    if ((s1->beg < s2->end)     /* if the two current intervals */
    &&  (s1->end > s2->beg)) {  /* overlap (intersection not empty) */
      d->beg = (s1->beg > s2->beg) ? s1->beg : s2->beg;
      d->end = (s1->end < s2->end) ? s1->end : s2->end;
      break;                    /* set the first interval */
    }                           /* and abort the search loop */
    if (s1->end < s2->end) { if ((++s1)->beg >= TRN_END) break; }
    else                   { if ((++s2)->beg >= TRN_END) break; }
  }                             /* skip interval ending earlier */
  if ((s1->beg >= TRN_END) || (s2->beg >= TRN_END)) {
    d->beg = d->end = TRN_END; dst->all = 0.0; dst->off = 0; return; }
  while (1) {                   /* trans. id list intersection loop */
    if ((s1->beg < s2->end)     /* if the two current intervals */
    &&  (s1->end > s2->beg)) {  /* overlap (intersection not empty) */
      a = (s1->beg > s2->beg) ? s1->beg : s2->beg;
      z = (s1->end < s2->end) ? s1->end : s2->end;
      if (a > d->end) {         /* if the intervals are disjoint */
        supp += d->end -d->beg; /* add interval length to support */
        (++d)->beg = a;         /* and start a new output interval */
      }
      d->end = z;               /* set/update the interval end */
    }
    if (s1->end < s2->end) { if ((++s1)->end >= TRN_END) break; }
    else                   { if ((++s2)->end >= TRN_END) break; }
  }                             /* skip interval ending earlier */
  supp += d->end -d->beg; d++;  /* add last output interval */
  d->beg = d->end = TRN_END;    /* store a sentinel at the list end */
  dst->all = (SUPP)(supp/width);/* compute and store the support */
  dst->off = (size_t)(++d -dst->ivals);
}  /* xisect() */               /* note offset to union list */

/*--------------------------------------------------------------------*/

static size_t xmerge (XIVLIST *dst, XIVLIST *src1, XIVLIST *src2,
                      double width)
{                               /* --- merge two interval lists */
  IVAL   *s1, *s2, *d;          /* to traverse sources and dest. */
  double supp;                  /* support of the merge/union */

  assert(dst && src1 && src2    /* check the function arguments */
  &&    (src1->ivals[0].beg < TRN_END)
  &&    (src2->ivals[0].beg < TRN_END) && (width >= 0));
  s1 = src1->ivals +src1->off;  /* get sources and destination */
  s2 = src2->ivals +src2->off;  /* (both refer to the union lists) */
  d  = dst->ivals  +dst->off; supp = 0.0;
  *d = (s1->beg < s2->beg) ? *s1 : *s2;
  if (d->beg >= TRN_END) { dst->one = 0.0; return 0; }
  while (1) {                   /* trans. id list intersection loop */
    if (s1->beg <= s2->beg) {   /* if source 1 has earlier interval */
      if (s1->beg >= TRN_END)   /* if both lists are exhausted, */
        break;                  /* abort the union loop */
      if      (s1->beg > d->end) { /* if the intervals are disjoint */
        supp += d->end -d->beg; /* add interval length to support */
        *++d = *s1; }           /* and start a new output interval */
      else if (s1->end > d->end)/* if the intervals overlap */
        d->end = s1->end;       /* update the interval end */
      s1++; }                   /* set/update the interval end */
    else {                      /* if source 1 has earlier interval */
      if      (s2->beg > d->end) { /* if the intervals are disjoint */
        supp += d->end -d->beg; /* add interval length to support */
        *++d = *s2; }           /* and start a new output interval */
      else if (s2->end > d->end)/* if the intervals overlap */
        d->end = s2->end;       /* update the interval end */
      s2++;                     /* set/update the interval end */
    }                           /* and skip the porcessed interval */
  }
  supp += d->end -d->beg; d++;  /* add last output interval */
  d->beg = d->end = TRN_END;    /* store a sentinel at the list end */
  dst->one = (SUPP)(supp/width);/* compute and store the support */
  return (size_t)(d-dst->ivals);/* return the number of intervals */
}  /* xmerge() */

#endif /* #ifndef CCN_OVERLAP .. #else .. */
/*----------------------------------------------------------------------
  Common Functions
----------------------------------------------------------------------*/

static int clomax (COCONAD *ccn, TRAIN **trains, ITEM n,
                   TRAIN **excl, ITEM e, SUPP supp, SUPP xmax)
{                               /* --- check for a closed/maximal set */
  SUPP s;                       /* support of an item set */

  assert(ccn && trains          /* check the function arguments */
  &&    (n >= 0) && (excl || (e <= 0)) && (supp >= 0) && (xmax >= 0));
  if (ccn->target & CCN_MAXIMAL)/* if to check for a maximal set, */
    supp = ccn->supp;           /* get the minimum support threshold */
  if (xmax >= supp)             /* if frequent or perfect extension, */
    return 0;                   /* abort (set is not closed/maximal) */
  while (--e >= 0) {            /* traverse the excluded items and */
    trains[n] = excl[e];        /* compute support of extended sets */
    s = support(trains, n+1, ccn->width, ccn->heap);
    #ifdef CCN_OVERLAP          /* if influence map overlap support */
    if (ccn->mix > 0)           /* if combined support mode */
      s = (SUPP)(((1.0-ccn->mix) *(double)s) +ccn->mix
        * (double)intsupp(trains, n+1, ccn->width, ccn->heap));
    #endif                      /* add standard (MIS) support */
    if (s >= supp) return 0;    /* if frequent or perfect extension, */
  }                             /* abort (set is not closed/maximal) */
  return -1;                    /* return that set is closed/maximal */
}  /* clomax() */

/*--------------------------------------------------------------------*/
#ifndef CCN_OVERLAP             /* if standard (MIS) support */

static int seqclomax (COCONAD *ccn, TRAIN **trains, ITEM n,
                      TRAIN **exts, ITEM e, SUPP supp, SUPP xmax)
{                               /* --- check for closed/maximal seq. */
  ITEM i, k;                    /* loop variables for extensions */

  assert(ccn                    /* check the function arguments */
  &&  trains && (n >= 0) && (exts || (e <= 0)) && (supp >= 0));
  if (ccn->target & CCN_MAXIMAL)/* if to check for a maximal seq., */
    supp = ccn->supp;           /* get the minimum support threshold */
  if (xmax >= supp)             /* if frequent or perfect extension, */
    return 0;                   /* abort (set is not closed/maximal) */
  if (ccn->target & CCN_APPEND) /* if only to consider appending, */
    return -1;                  /* all checks are done, so success */
  for (i = 0; i < e; i++) {     /* traverse the possible extensions */
    if (!exts[i]) continue;     /* skip the already contained items */
    memcpy(ccn->excl, trains, (size_t)n *sizeof(TRAIN*));
    #if 0                       /* not needed, already checked: */
    ccn->excl[n] = exts[i];     /* copy trains and append extension */
    if (seqsupp(ccn->excl, n+1, ccn->width, ccn->heap) >= supp)
      return 0;                 /* if frequent or perfect extension, */
    #endif                      /* abort (seq. is not closed/maximal) */
    for (k = n; --k >= 0; ) {   /* traverse the insertion positions */
      ccn->excl[k+1] = ccn->excl[k];
      ccn->excl[k]   = exts[i]; /* build extended item sequence */
      if (seqsupp(ccn->excl, n+1, ccn->width, ccn->heap) >= supp)
        return 0;               /* compute support of extended seqs.: */
    }                           /* if frequent or perfect extension, */
  }                             /* abort (seq. is not closed/maximal) */
  return -1;                    /* return that set is closed/maximal */
}  /* seqclomax() */

/*--------------------------------------------------------------------*/

static int pexcube (COCONAD *ccn, TRAIN **trains, ITEM n,
                    ITEM e, ITEM x, SUPP supp)
{                               /* --- traverse perf. exts. hypercube */
  int  r = 0;                   /* result of functions/error status */
  ITEM k;                       /* (total) number of excluded items */
  SUPP s, m = 0;                /* (maximum) support (of extension) */

  assert(ccn && trains          /* check the function arguments */
  &&    (n >= 0) && (e >= 0) && (x >= 0) && (supp >= 0));
  #ifdef CCN_ABORT              /* if to check for interrupt */
  if (sig_aborted()) return -1; /* if execution was aborted, */
  #endif                        /* abort the recursion */
  if (supp <= ccn->limit) {     /* if the support limit is reached, */
    while (--x >= 0) {          /* traverse all remaining extensions */
      isr_addpex(ccn->report, trn_item(ccn->pexs[x]));
      trains[n++] = ccn->pexs[x];  /* add them as perfect extensions */
    }                              /* and to the collected trains */
    if (!(ccn->target & (CCN_CLOSED|CCN_MAXIMAL))
    ||  clomax(ccn, trains, n, ccn->excl, e, supp, 0))
      return isr_report(ccn->report);
    return 0;                   /* report the extended item set */
  }                             /* and abort the function */
  if (isr_xable(ccn->report,1)){/* if an item can be added */
    for (k = e; --x >= 0; ) {   /* traverse the possible extensions */
      trains[n] = ccn->pexs[x]; /* add next extension to trains */
      s = support(trains, n+1, ccn->width, ccn->heap);
      #ifdef CCN_OVERLAP        /* if influence map overlap support */
      if (ccn->mix > 0)         /* if combined support mode */
        s = (SUPP)(((1.0-ccn->mix) *(double)s) +ccn->mix
          * (double)intsupp(trains, n+1, ccn->width, ccn->heap));
      #endif                    /* add standard (MIS) support */
      if (s < ccn->supp)        /* compute the item set support */
        continue;               /* and skip infrequent items */
      if (s > m) m = s;         /* find maximum extension support */
      r = isr_addnc(ccn->report, trn_item(trains[n]), s);
      if (r < 0) return r;      /* add current item to the reporter */
      r = pexcube(ccn, trains, n+1, k, x, s);
      isr_remove(ccn->report,1);/* recursively find freq. item sets, */
      if (r < 0) return r;         /* then remove the current item */
      ccn->excl[k++] = trains[n];  /* from the item set reporter and */
    }                              /* add it to the excluded items */
  }                                /* (for later closedness check) */
  if (!(ccn->target & (CCN_CLOSED|CCN_MAXIMAL))
      ||  clomax(ccn, trains, n, ccn->excl, e, supp, m))
    r = isr_report(ccn->report);/* report the current item set */
  return r;                     /* return the error status */
}  /* pexcube() */

/*--------------------------------------------------------------------*/

static void ccnclean (COCONAD *ccn)
{                               /* --- clean up recursion data */
  if (ccn->cmf)  cmf_delete(ccn->cmf);
  if (ccn->cmx)  cmx_delete(ccn->cmx);
  if (ccn->used) free(ccn->used);
}  /* ccnclean() */             /* delete repositories and array */

/*--------------------------------------------------------------------*/

static int report2cmx (COCONAD *ccn, SUPP xmax)
{                               /* --- report to prefix tree chain */
  RSUPP s = isr_supp(ccn->report);  /* get the permutation support */
  ITEM  n = isr_cnt (ccn->report);  /* and check the sequence size */
  if ((xmax < (ccn->target & CCN_MAXIMAL) ? ccn->supp : (SUPP)s)
  &&  (n >= isr_zmin(ccn->report)) && (n <= isr_xmax(ccn->report)))
    return cmx_update(ccn->cmx, isr_items(ccn->report), n, s);
  return 0;                     /* update repository with sequence */
}  /* report2cmx() */

#endif  /* #ifndef CCN_OVERLAP */
/*--------------------------------------------------------------------*/

static int report2cmf (COCONAD *ccn, SUPP xmax)
{                               /* --- report to list repository */
  RSUPP s = isr_supp(ccn->report);  /* get the permutation support */
  ITEM  n = isr_cnt (ccn->report);  /* and check the sequence size */
  if ((xmax < (ccn->target & CCN_MAXIMAL) ? ccn->supp : (SUPP)s)
  &&  (n >= isr_zmin(ccn->report)) && (n <= isr_xmax(ccn->report)))
    return cmf_add(ccn->cmf, isr_items(ccn->report), n, s);
  return 0;                     /* add sequence to repository */
}  /* report2cmf() */

/*----------------------------------------------------------------------
  CoCoNAD (basic version without point and train filtering)
----------------------------------------------------------------------*/

static int rec_basic (COCONAD *ccn,
                      TRAIN **trains, ITEM n, ITEM p, ITEM e,
                      TRAIN **exts,   ITEM x, SUPP supp)
{                               /* --- recursive part of coconad */
  int  r = 0;                   /* result of functions/error status */
  ITEM k = e;                   /* (total) number of excluded items */
  SUPP pex;                     /* minimum support for perfect exts. */
  SUPP s, m = 0;                /* (maximum) support (of extension) */
  #ifndef CCN_OVERLAP           /* if integer (MIS) support type */
  ITEM i;                       /* loop variable for items */
  #endif                        /* (needed for perfect extensions) */

  assert(ccn && trains          /* check the function arguments */
  &&    (n >= 0) && (p >= 0) && (e >= 0)
  &&    exts && (x >= 0) && (supp >= 0));
  #ifdef CCN_ABORT              /* if to check for interrupt */
  if (sig_aborted()) return -1; /* if execution was aborted, */
  #endif                        /* abort the recursion */
  if ((x > 0)                   /* if there are extension items */
  &&  isr_xable(ccn->report,1)){/* and another item can be added */
    pex = (ccn->mode & CCN_PERFECT)
        ? supp : SUPP_MAX;      /* get support for perfect extensions */
    while (--x >= 0) {          /* traverse the possible extensions */
      trains[n] = exts[x];      /* add next extension to trains */
      s = support(trains, n+1, ccn->width, ccn->heap);
      #ifdef CCN_OVERLAP        /* if influence map overlap support */
      if (ccn->mix > 0)         /* if combined support mode */
        s = (SUPP)(((1.0-ccn->mix) *(double)s) +ccn->mix
          * (double)intsupp(trains, n+1, ccn->width, ccn->heap));
      #endif                    /* add standard (MIS) support */
      if (s < ccn->supp)        /* compute the item set support, */
        continue;               /* skip infrequent items, and */
      if (s >= pex) {           /* collect perfect extensions */
        #ifdef CCN_OVERLAP      /* if influence map overlap support */
        isr_addpex(ccn->report, trn_item(trains[n])); continue; }
        #else                   /* if integer (MIS) support type */
        ccn->pexs[p++] = trains[n]; continue; }
        #endif                  /* treat perfect exts. differently */
      if (s > m) m = s;         /* find maximum extension support */
      r = isr_addnc(ccn->report, trn_item(trains[n]), s);
      if (r < 0) return r;      /* add current item to the reporter */
      r = rec_basic(ccn, trains, n+1, p, k, exts, x, s);
      isr_remove(ccn->report,1);/* recursively find freq. item sets */
      if (r < 0) return r;      /* remove current item from reporter */
      ccn->excl[k++] = trains[n];
    }                           /* add item to the excluded items */
  }                             /* (for later closedness check) */
  #ifndef CCN_OVERLAP           /* if integer (MIS) support type */
  if (p > 0) {                  /* if there are perfect extensions */
    memcpy(trains+n, ccn->pexs, (size_t)p *sizeof(TRAIN*));
    ccn->limit = support(trains, n+p, ccn->width, ccn->heap);
    if (ccn->limit < supp) {    /* compute support of whole item set */
      if (m >= ccn->limit) e = k;/* and if it is lower than expected */
      return pexcube(ccn, trains, n, e, p, supp);
    }                           /* traverse perfect ext. hypercube */
    for (i = 0; i < p; i++)     /* add perfect exts. to the reporter */
      isr_addpex(ccn->report, trn_item(ccn->pexs[i]));
    e = k; m = 0;               /* adapt number of excluded items */
  }                             /* and maximal extension support */
  #endif
  if (!(ccn->target & (CCN_CLOSED|CCN_MAXIMAL))
  ||  clomax(ccn, trains, n+p, ccn->excl, e, supp, m))
    r = isr_report(ccn->report);/* report the current item set */
  return r;                     /* return the error status */
}  /* rec_basic() */

/*--------------------------------------------------------------------*/

int ccn_basic (COCONAD *ccn)
{                               /* --- closed cont. neuron ass. dect. */
  int   r = 0;                  /* result of recursion/error status */
  ITEM  i, k, n;                /* loop variable, number of items */
  TRAIN *t, **trains;           /* to traverse the trains */

  assert(ccn);                  /* check the function arguments */
  if (ccn->supp > (SUPP)tns_maxlen(ccn->trnset))
    return 0;                   /* check the maximum train length */
  k = tns_cnt(ccn->trnset);     /* get the number of items and */
  if (k <= 0) return 0;         /* check for at least one item */
  trains = (TRAIN**)malloc(4*(size_t)k *sizeof(TRAIN*)
                            +(size_t)k *sizeof(double*));
  if (!trains) return -1;       /* create the work arrays and */
  ccn->pexs = trains +k+k;      /* organize the memory, that is, */
  ccn->excl = trains +k+k+k;    /* store the different train arrays */
  ccn->heap = (const double**)(trains+4*k);
  for (i = n = 0; i < k; i++) {    /* traverse the trains and */
    t = tns_train(ccn->trnset, i); /* collect the relevant ones */
    if ((SUPP)trn_len(t) >= ccn->supp) trains[n++] = t;
  }                             /* (sufficiently many points) */
  if (n > 0) {                  /* if there are frequent items */
    ptr_qsort(trains, (size_t)n, -1, trn_cmpsz, NULL);
    r = rec_basic(ccn, trains+n, 0, 0, 0, trains, n, SUPP_MAX);
  }                             /* recursively find freq. item sets */
  free(trains);                 /* delete the allocated arrays */
  return (r < 0) ? r : 0;       /* return the error status */
}  /* ccn_basic() */

/*----------------------------------------------------------------------
  CoCoNAD (refined version with point and train filtering)
----------------------------------------------------------------------*/

static int rec_filter (COCONAD *ccn, TRAINDB *tdb, ITEM n,
                       ITEM p, ITEM e, ITEM x, SUPP supp)
{                               /* --- recursive part of coconad */
  int     r = 0;                /* result of functions/error status */
  ITEM    k = e, l, y;          /* number of (excluded) items */
  size_t  z;                    /* size of conditional database */
  SUPP    pex;                  /* minimum support for perfect exts. */
  SUPP    s, m = 0;             /* current and maximum support */
  TRAINDB *proj = NULL;         /* projected/filtered train database */
  void    *d;                   /* pointer to free destination memory */

  assert(ccn && tdb && (n >= 0) /* check the function arguments */
  &&    (p >= 0) && (e >= 0) && (x >= 0) && (supp >= 0));
  #ifdef CCN_ABORT              /* if to check for interrupt */
  if (sig_aborted()) return -1; /* if execution was aborted, */
  #endif                        /* abort the recursion */
  if ((x > 0)                   /* if there are extension items */
  &&  isr_xable(ccn->report,1)){/* and another item can be added */
    z = (size_t)n+2+(size_t)p +(size_t)x-1 +(size_t)e+(size_t)x;
    z += z & 1;                 /* compute maximal number of trains */
    proj = (TRAINDB*)malloc(sizeof(TRAINDB)
                          +(size_t)ccn->max *sizeof(IVAL)
                          +z *sizeof(TRAIN*) +tdb->size);
    if (!proj) return -1;       /* allocate memory for projection */
    proj->trains = (TRAIN**)(proj->ivals +ccn->max +1);
    proj->exts   = proj->trains +n +2 +p;
    proj->excl   = proj->exts   +x -1;
    proj->mem    = (void*)(proj->trains +z);
    pex = (ccn->mode & CCN_PERFECT)
        ? supp : SUPP_MAX;      /* get support for perfect extensions */
    if ((x > 4)                 /* if there are enough trains left, */
    &&  (ccn->mode & CCN_REORDER))  /* re-sort them by their length */
      ptr_qsort(tdb->exts, (size_t)x, -1, trn_cmpsz, NULL);
    while (--x >= 0) {          /* traverse the extensions */
      s = (SUPP)trn_isect(proj->ivals, tdb->ivals,
                          tdb->exts[x], fabs(ccn->width));
      if (s < ccn->supp)        /* intersect the interval list and */
        continue;               /* check the number of used spikes */
      tdb->trains[n] = tdb->exts[x];    /* add the extension train */
      d = proj->mem;            /* and filter the collected trains */
      y = tna_filter(proj->trains, tdb->trains, n+1,
                     proj->ivals, ccn->supp, &d);
      if (y < n+1) continue;    /* if one becomes empty, skip ext. */
      s = support(proj->trains, n+1, ccn->width, ccn->heap);
      #ifdef CCN_OVERLAP        /* if influence map overlap support */
      if (ccn->mix > 0)         /* if combined support mode */
        s = (SUPP)(((1.0-ccn->mix) *(double)s) +ccn->mix
          * (double)intsupp(proj->trains, n+1, ccn->width, ccn->heap));
      #endif                    /* add standard (MIS) support */
      if (s < ccn->supp)        /* compute the item set support, */
        continue;               /* skip infrequent items, and */
      if (s >= pex) {           /* collect perfect extensions */
        #ifdef CCN_OVERLAP      /* if influence map overlap support */
        isr_addpex(ccn->report, trn_item(tdb->trains[n])); continue; }
        #else                   /* if integer (MIS) support type */
        ccn->pexs[p++] = tdb->trains[n]; continue; }
        #endif                  /* treat perfect exts. differently */
      if (s > m) m = s;         /* find maximum extension support */
      r = isr_addnc(ccn->report, trn_item(tdb->trains[n]), s);
      if (r < 0) break;         /* add current item to the reporter */
      l = tna_reduce(proj->excl,  tdb->excl, k,   /* filter/reduce */
                     proj->ivals, ccn->supp, &d); /* excluded trains */
      y = tna_reduce(proj->exts,  tdb->exts, x,   /* filter/reduce */
                     proj->ivals, ccn->supp, &d); /* extension trains */
      proj->size = (size_t)((char*)d -(char*)proj->mem);
      r = rec_filter(ccn, proj, n+1, p, l, y, s);
      isr_remove(ccn->report,1);/* recursively find freq. item sets, */
      if (r < 0) break;         /* then remove the current item from */
      tdb->excl[k++] = tdb->trains[n];  /* the item set reporter and */
    }                           /* add it to the excluded items */
    free(proj);                 /* deallocate the projection memory */
    if (r < 0) return r;        /* if an error occurred, abort */
  }
  #ifndef CCN_OVERLAP           /* if integer (MIS) support type */
  if (p > 0) {                  /* if there are perfect extensions */
    memcpy(tdb->trains+n, ccn->pexs, (size_t)p *sizeof(TRAIN*));
    ccn->limit = support(tdb->trains, n+p, ccn->width, ccn->heap);
    if (ccn->limit < supp) {    /* compute support of whole item set */
      if (m >= ccn->limit) e = k;  /* if it is lower than expected, */
      ccn->excl = tdb->excl;       /* traverse perfect ext. hypercube */
      return pexcube(ccn, tdb->trains, n, e, p, supp);
    }                           /* (extensions have to be redone) */
    for (l = 0; l < p; l++)     /* add perfect exts to the reporter */
      isr_addpex(ccn->report, trn_item(ccn->pexs[l]));
    e = k; m = 0;               /* adapt number of excluded items */
  }                             /* and maximal extension support */
  #endif
  if (!(ccn->target & (CCN_CLOSED|CCN_MAXIMAL))
  ||  clomax(ccn, tdb->trains, n+p, tdb->excl, e, supp, m))
    r = isr_report(ccn->report);/* report the current item set */
  return r;                     /* return the error status */
}  /* rec_filter() */

/*--------------------------------------------------------------------*/

int ccn_filter (COCONAD *ccn)
{                               /* --- closed cont. neuron ass. dect. */
  int     r;                    /* result of recursion/error status */
  ITEM    i, k, n;              /* loop variables, number of items */
  TID     s;                    /* size/length of a train */
  TRAIN   *t, **trains;         /* to traverse the trains */
  TRAINDB *tdb;                 /* train database for recursion */
  IVAL    *ival;                /* to initialize the intervals */

  assert(ccn);                  /* check the function arguments */
  if (ccn->supp > (SUPP)tns_maxlen(ccn->trnset))
    return 0;                   /* check the maximum train length */
  k = tns_cnt(ccn->trnset);     /* get the number of items and */
  if (k <= 0) return 0;         /* check for at least one item */
  tdb = (TRAINDB*)malloc(sizeof(TRAINDB) +sizeof(IVAL)
                         +4*(size_t)k    *sizeof(TRAIN*)
                         +  (size_t)k    *sizeof(double*));
  if (!tdb) return -1;          /* allocate the work memory */
  trains      = (TRAIN**)(tdb->ivals+2);
  ccn->heap   = (const double**)(trains+4*k);
  ccn->pexs   = trains;         /* organize the memory, that is, */
  tdb->trains = trains += k;    /* store the different train arrays */
  tdb->exts   = trains += k;
  tdb->excl   = trains +  k;
  tdb->mem    = NULL;
  tdb->size   = tna_dstsize(k, tns_extent(ccn->trnset));
  ival = tdb->ivals;   ival->beg = -INFINITY; ival->end = INFINITY;
  ival = tdb->ivals+1; ival->beg =  INFINITY; ival->end = INFINITY;
  for (ccn->max = 0, i = n = 0; i < k; i++) {
    t = tns_train(ccn->trnset, i);    /* traverse the trains, but */
    s = trn_len(t);             /* skip trains that are too short */
    if ((SUPP)s < ccn->supp) continue;
    if (s > ccn->max) ccn->max = s;  /* find the maximum train size */
    trains[n++] = t;            /* collect sufficiently long trains */
  }                             /* (extensions of empty set) */
  r = (n <= 0) ? 0              /* recursively find freq. item sets */
    : rec_filter(ccn, tdb, 0, 0, 0, n, SUPP_MAX);
  free(tdb);                    /* deallocate the work memory */
  return (r < 0) ? r : 0;       /* return the error status */
}  /* ccn_filter() */

/*----------------------------------------------------------------------
  CoCoNAD (closed item set filtering with repository)
----------------------------------------------------------------------*/

#ifdef CCN_OVERLAP              /* if influence map overlap support */
static int rec_cmrepo (COCONAD *ccn, TRAINDB *tdb, ITEM n, ITEM x,
                       SUPP supp)
#else                           /* if standard (MIS) support */
static int rec_cmrepo (COCONAD *ccn, TRAINDB *tdb, ITEM n, ITEM x)
#endif                          /* (do not need parent support) */
{                               /* --- recursive part of coconad */
  int     r = 0;                /* result of functions/error status */
  ITEM    k;                    /* item buffer, filtering result */
  SUPP    s;                    /* support of extended item set */
  size_t  z;                    /* size of conditional database */
  TRAINDB *proj = NULL;         /* projected/filtered train database */
  void    *d;                   /* pointer to free destination memory */
  #ifdef CCN_OVERLAP            /* if influence map overlap support */
  SUPP    pex;                  /* support for a perfect extension */
  #endif                        /* (not possible for MIS support) */

  assert(ccn && tdb             /* check the function arguments */
  &&    (n >= 0) && (x >= 0));
  #ifdef CCN_ABORT              /* if to check for interrupt */
  if (sig_aborted()) return -1; /* if execution was aborted, */
  #endif                        /* abort the recursion */
  z = (size_t)n+2+(size_t)x;    /* compute max. number of trains */
  z += z & 1;                   /* (even number for alignment) */
  proj = (TRAINDB*)malloc(sizeof(TRAINDB)
                          +(size_t)ccn->max *sizeof(IVAL)
                          +z *sizeof(TRAIN*) +tdb->size);
  if (!proj) return -1;         /* allocate memory for projection */
  proj->trains = (TRAIN**)(proj->ivals +ccn->max +1);
  proj->exts   = proj->trains +n +2;
  proj->mem    = (void*)(proj->trains +z);
  #ifdef CCN_OVERLAP            /* if influence map overlap support */
  pex = (ccn->mode & CCN_PERFECT) ? supp : SUPP_MAX;
  #endif                        /* get support for perfect extensions */
  while (--x >= 0) {            /* traverse the extensions */
    s = (SUPP)trn_isect(proj->ivals, tdb->ivals,
                        tdb->exts[x], fabs(ccn->width));
    if (s < ccn->supp)          /* intersect the interval list and */
      continue;                 /* check the number of used spikes */
    tdb->trains[n] = tdb->exts[x];      /* add the extension train */
    d = proj->mem;              /* and filter the collected trains */
    k = tna_filter(proj->trains, tdb->trains, n+1,
                   proj->ivals, ccn->supp, &d);
    if (k < 0) continue;        /* if one becomes empty, skip ext. */
    s = support(proj->trains, n+1, ccn->width, ccn->heap);
    #ifdef CCN_OVERLAP          /* if influence map overlap support */
    if (ccn->mix > 0)           /* if combined support mode */
      s = (SUPP)(((1.0-ccn->mix) *(double)s) +ccn->mix
        * (double)intsupp(proj->trains, n+1, ccn->width, ccn->heap));
    #endif                      /* add standard (MIS) support */
    if (s < ccn->supp)          /* compute the item set support */
      continue;                 /* and skip infrequent items */
    k = trn_item(tdb->trains[n]);  /* get the extension item */
    #ifdef CCN_OVERLAP          /* if influence map overlap support */
    if (s >= pex) {             /* collect perfect extensions */
      isr_addpex(ccn->report, k); continue; }
    #endif                      /* (not possible with MIS support) */
    r = isr_addnc(ccn->report, k, s);
    if (r <  0) break;          /* add current item to the reporter */
    if (r <= 0) continue;       /* check if item needs processing */
    if (isr_xable(ccn->report, 1)) { /* if another item can be added */
      k = tna_reduce(proj->exts, tdb->exts, x,    /* filter/reduce */
                     proj->ivals, ccn->supp, &d); /* extension trains */
      if (k > 0) {              /* if there are extension items */
        proj->size = (size_t)((char*)d -(char*)proj->mem);
        #ifdef CCN_OVERLAP      /* if influence map overlap support */
        r = rec_cmrepo(ccn, proj, n+1, k, s);
        #else                   /* if standard (MIS) support */
        r = rec_cmrepo(ccn, proj, n+1, k);
        #endif                  /* do not need parent support */
        if (r < 0) break;       /* recursively find freq. item sets */
      }                         /* in the created projection */
    }                           /* and check for a search error */
    if (ccn->repfn) ccn->repfn(ccn, -1);
    else isr_report(ccn->report);  /* report the current item set */
    isr_remove(ccn->report, 1); /* remove current item from reporter */
    if (r < 0) break;           /* and check for a reporting error */
  }                             /* from the item set reporter */
  free(proj);                   /* deallocate the projection memory */
  return r;                     /* return the error status */
}  /* rec_cmrepo() */

/*--------------------------------------------------------------------*/

int ccn_cmrepo (COCONAD *ccn)
{                               /* --- closed cont. neuron ass. dect. */
  int     r;                    /* result of recursion/error status */
  ITEM    i, k, n;              /* loop variable, number of items */
  TID     s;                    /* size/length of a train */
  TRAIN   *t, **trains;         /* to traverse the trains */
  TRAINDB *tdb;                 /* train database for recursion */
  IVAL    *ival;                /* to initialize the intervals */

  assert(ccn);                  /* check the function arguments */
  if (ccn->supp > (SUPP)tns_maxlen(ccn->trnset))
    return 0;                   /* check the maximum train length */
  k = tns_cnt(ccn->trnset);     /* get the number of items and */
  if (k <= 0) return 0;         /* check for at least one item */
  if (!(ccn->mode & CCN_REPOLIST)
  ||  !(ccn->target & (CCN_CLOSED|CCN_MAXIMAL))) {
    ccn->cmf = NULL;            /* if frequent sets, no repository */
    ccn->repfn = (REPORTFN*)0;} /* and no pattern reporting function */
  else {                        /* if to filter closed/max. sequences */
    ccn->cmf = cmf_create((size_t)(1024*1024));
    if (!ccn->cmf) return -1;   /* create a closed/maximal filter */
    ccn->repfn = report2cmf;    /* to collect the result sequences */
  }                             /* and get the reporting function */
  tdb = (TRAINDB*)malloc(sizeof(TRAINDB) +sizeof(IVAL)
                         +2*(size_t)k    *sizeof(TRAIN*)
                         +  (size_t)k    *sizeof(double*));
  if (!tdb) return -1;          /* allocate the work memory */
  tdb->trains = trains = (TRAIN**)(tdb->ivals+2);
  tdb->exts   = trains += k;    /* get and organize memory, that is, */
  tdb->mem    = NULL;           /* store the different train arrays */
  tdb->size   = tna_dstsize(k, tns_extent(ccn->trnset));
  ccn->heap   = (const double**)(trains +k);
  ival = tdb->ivals;   ival->beg = -INFINITY; ival->end = INFINITY;
  ival = tdb->ivals+1; ival->beg =  INFINITY; ival->end = INFINITY;
  for (ccn->max = 0, i = n = 0; i < k; i++) {
    t = tns_train(ccn->trnset, i);    /* traverse the trains, but */
    s = trn_len(t);             /* skip trains that are too short */
    if ((SUPP)s < ccn->supp) continue;
    if (s > ccn->max) ccn->max = s;
    trains[n++] = t;            /* find the maximum train size and */
  }                             /* collect the relevant trains */
  r = (n <= 0) ? 0              /* recursively find freq. item sets */
    #ifdef CCN_OVERLAP          /* if influence map overlap support */
    : rec_cmrepo(ccn, tdb, 0, n, SUPP_MAX);
    #else                       /* if standard (MIS) support */
    : rec_cmrepo(ccn, tdb, 0, n);
    #endif                      /* (do not need parent support) */
  free(tdb);                    /* deallocate the work memory */
  if ((r >= 0) && ccn->cmf) {   /* if to report from the c/m filter */
    r = (ccn->target & CCN_MAXIMAL)
      ? cmf_maximal(ccn->cmf, ccn->report)
      : cmf_closed (ccn->cmf, ccn->report);
    cmf_delete(ccn->cmf);       /* report closed/maximal sequences, */
  }                             /* delete the closed/maximal filter */
  return (r < 0) ? r : 0;       /* return the error status */
}  /* ccn_cmrepo() */

/* Filtering closed and maximal item sets with a repository makes  */
/* perfect extension pruning impossible, because here the perfect  */
/* extensions are not necessarily true perfect extensions and may  */
/* rather cause a reordering of the items, which is not compatible */
/* with a repository (the repository may be pruned incorrectly).   */

/*----------------------------------------------------------------------
  CoCoNAD (item sequences, basic version)
----------------------------------------------------------------------*/
#ifndef CCN_OVERLAP             /* if standard (MIS) support */

static int rec_seqbasic (COCONAD *ccn, TRAIN **trains, ITEM n,
                         TRAIN **exts, ITEM x, SUPP supp)
{                               /* --- recursive part of coconad */
  int  r = 0;                   /* result of functions/error status */
  ITEM i;                       /* loop variable for extensions */
  SUPP s, m = 0;                /* (maximum) support (of extension) */

  assert(ccn && trains          /* check the function arguments */
  &&    (n >= 0) && exts && (x >= 0) && (supp >= 0));
  #ifdef CCN_ABORT              /* if to check for interrupt */
  if (sig_aborted()) return -1; /* if execution was aborted, */
  #endif                        /* abort the recursion */
  if ((x-n > 0)                 /* if there are extension items */
  &&  isr_xable(ccn->report,1)){/* and another item can be added */
    for (i = 0; i < x; i++) {   /* traverse the possible extensions */
      if (!exts[i]) continue;   /* skip removed extensions (perms.) */
      trains[n] = exts[i];      /* add next extension to trains */
      s = seqsupp(trains, n+1, ccn->width, ccn->heap);
      if (s < ccn->supp)        /* compute the sequence support, */
        continue;               /* skip infrequent item sequences */
      if (s > m) m = s;         /* find maximum extension support */
      r = isr_addnc(ccn->report, trn_item(trains[n]), s);
      if (r < 0) return r;      /* add current item to the reporter */
      if (ccn->target & CCN_PERMUT)   /* if to find permutations, */
        exts[i] = NULL;         /* (temporarily) remove its train */
      r = rec_seqbasic(ccn, trains, n+1, exts, x, s);
      exts[i] = trains[n];      /* recursively find freq. item seqs., */
      isr_remove(ccn->report,1);/* restore the possibly removed train */
      if (r < 0) return r;      /* and remove the current item */
    }                           /* from the item set reporter */
  }
  if (!(ccn->target & (CCN_CLOSED|CCN_MAXIMAL))
  ||  seqclomax(ccn, trains, n, exts, x, supp, m))
    r = isr_report(ccn->report);/* report the current item set */
  return r;                     /* return the error status */
}  /* rec_seqbasic() */

/*--------------------------------------------------------------------*/

int ccn_seqbasic (COCONAD *ccn)
{                               /* --- closed cont. neuron seq. dect. */
  int   r = 0;                  /* result of recursion/error status */
  ITEM  i, k, n;                /* loop variable, number of items */
  TRAIN *t, **trains;           /* to traverse the trains */

  assert(ccn);                  /* check the function arguments */
  if (ccn->supp > (SUPP)tns_maxlen(ccn->trnset))
    return 0;                   /* check the maximum train length */
  k = tns_cnt(ccn->trnset);     /* get the number of items */
  if (k <= 0) return 0;         /* and the maximum pattern length */
  n = (ccn->target & CCN_SEQUENCE)
    ? tns_maxseq(ccn->trnset, fabs(ccn->width)) +1 : k;
  if (n < k) n = k;             /* ensure at least number of items */
  trains = (TRAIN**)malloc((size_t)(k+k+n) *sizeof(TRAIN*)
                          +(size_t) n      *sizeof(double*));
  if (!trains) return -1;       /* create the work arrays */
  ccn->excl = trains +k+n;      /* and organize the memory */
  ccn->heap = (const double**)(trains+k+k+n);
  for (i = n = 0; i < k; i++) { /* traverse the trains */
    t = tns_train(ccn->trnset, i);
    if ((SUPP)trn_len(t) >= ccn->supp) trains[n++] = t;
  }                             /* collect the relevant trains */
  r = (n <= 0) ? 0              /* recursively find freq. item seqs. */
    : rec_seqbasic(ccn, trains+n, 0, trains, n, SUPP_MAX);
  free(trains);                 /* delete the allocated arrays */
  return (r < 0) ? r : 0;       /* return the error status */
}  /* ccn_seqbasic() */

/*----------------------------------------------------------------------
  CoCoNAD (item permutations/sequences, point and train filtering)
----------------------------------------------------------------------*/

static int rec_pmtfilter (COCONAD *ccn, TRAINDB *tdb, ITEM n, ITEM x,
                          SUPP supp)
{                               /* --- recursive part of coconad */
  int     r = 0;                /* result of functions/error status */
  ITEM    k;                    /* loop variable for extensions */
  ITEM    e;                    /* number of extension items */
  size_t  z;                    /* size of conditional database */
  SUPP    s, m = 0;             /* (maximum) support (of extension) */
  TRAINDB *proj = NULL;         /* projected/filtered train database */
  void    *d;                   /* pointer to free destination memory */

  assert(ccn && tdb             /* check the function arguments */
  &&    (n >= 0) && (x >= 0) && (supp >= 0));
  #ifdef CCN_ABORT              /* if to check for interrupt */
  if (sig_aborted()) return -1; /* if execution was aborted, */
  #endif                        /* abort the recursion */
  if ((x > 0)                   /* if there are extension items */
  &&  isr_xable(ccn->report,1)){/* and another item can be added */
    z = (size_t)n+2 +(size_t)x; /* compute max. number of trains */
    z += z & 1;                 /* (even number for alignment) */
    proj = (TRAINDB*)malloc(sizeof(TRAINDB)
                            +(size_t)ccn->max *sizeof(IVAL)
                            +z *sizeof(TRAIN*) +tdb->size);
    if (!proj) return -1;       /* allocate memory for projection */
    proj->trains = (TRAIN**)(proj->ivals +ccn->max +1);
    proj->exts   = proj->trains +n+2;
    proj->mem    = (void*)(proj->trains +z);
    for (k = 0; k < x; k++) {   /* traverse possible extensions */
      s = (SUPP)trn_seqisect(proj->ivals, tdb->ivals,
                             tdb->exts[k], fabs(ccn->width));
      if (s < ccn->supp)        /* intersect the interval list and */
        continue;               /* check the number of used spikes */
      tdb->trains[n] = tdb->exts[k];    /* add the extension train */
      d = proj->mem;            /* and filter the collected trains */
      e = tna_seqfilt(proj->trains, tdb->trains, n+1, proj->ivals,
                      fabs(ccn->width), ccn->supp, &d);
      if (e < n+1) continue;    /* if one becomes empty, skip ext. */
      s = seqsupp(proj->trains, n+1, ccn->width, ccn->heap);
      if (s < ccn->supp)        /* compute the sequence support, */
        continue;               /* skip infrequent sequences, and */
      if (s > m) m = s;         /* find maximum extension support */
      r = isr_addnc(ccn->report, trn_item(tdb->trains[n]), s);
      if (r < 0) break;         /* add current item to the reporter */
      tdb->exts[k] = NULL;      /* and (temporarily) remove its train */
      /* filter/reduce the extension trains (reduce times/points) */
      e = tna_seqred(proj->exts, tdb->exts, x,
                     proj->ivals, fabs(ccn->width), ccn->supp,
                     ccn->target & (CCN_CLOSED|CCN_MAXIMAL), &d);
      proj->size = (size_t)((char*)d -(char*)proj->mem);
      r = rec_pmtfilter(ccn, proj, n+1, e, s);
      isr_remove(ccn->report,1);/* recursively find freq. item perms. */
      tdb->exts[k] = tdb->trains[n];
      if (r < 0) break;         /* remove current item from reporter */
    }                           /* and restore its train */
    free(proj);                 /* deallocate the projection memory */
    if (r < 0) return r;        /* if an error occurred, abort */
  }
  if (!(ccn->target & (CCN_CLOSED|CCN_MAXIMAL))
  ||  seqclomax(ccn, tdb->trains, n, tdb->exts, x, supp, m))
    r = isr_report(ccn->report);/* report the current permutation */
  return r;                     /* return the error status */
}  /* rec_pmtfilter() */

/*--------------------------------------------------------------------*/

static int rec_seqfilter (COCONAD *ccn, TRAINDB *tdb, ITEM n, ITEM x,
                          SUPP supp)
{                               /* --- recursive part of coconad */
  int     r = 0;                /* result of functions/error status */
  ITEM    k;                    /* loop variable for extensions */
  ITEM    i, u;                 /* extension item and its first use */
  ITEM    e;                    /* number of extension items */
  size_t  y, z;                 /* size of conditional database */
  SUPP    s, m = 0;             /* (maximum) support (of extension) */
  TRAINDB *proj = NULL;         /* projected/filtered train database */
  void    *d;                   /* pointer to free destination memory */

  assert(ccn && tdb             /* check the function arguments */
  &&    (n >= 0) && (x >= 0) && (supp >= 0));
  #ifdef CCN_ABORT              /* if to check for interrupt */
  if (sig_aborted()) return -1; /* if execution was aborted, */
  #endif                        /* abort the recursion */
  if ((x > 0)                   /* if there are extension items */
  &&  isr_xable(ccn->report,1)){/* and another item can be added */
    z = (size_t)n+2 +(size_t)x; /* compute max. number of trains */
    z += z & 1;                 /* (even number for alignment) */
    y = (ccn->target & (CCN_CLOSED|CCN_MAXIMAL))
      ? 2*(sizeof(TRAIN)  +(size_t)ccn->max *sizeof(double)) : 0;
    proj = (TRAINDB*)malloc(sizeof(TRAINDB)
                          +(size_t)ccn->max *sizeof(IVAL)
                          +z *sizeof(TRAIN*) +tdb->size +y);
    if (!proj) return -1;       /* allocate memory for projection */
    proj->trains = (TRAIN**)(proj->ivals +ccn->max +1);
    proj->exts   = proj->trains +n+2;
    proj->mem    = (void*)(proj->trains +z);
    for (k = 0; k < x; k++) {   /* traverse possible extensions */
      s = (SUPP)trn_seqisect(proj->ivals, tdb->ivals,
                             tdb->exts[k], fabs(ccn->width));
      if (s < ccn->supp)        /* intersect the interval list and */
        continue;               /* check the number of used spikes */
      tdb->trains[n] = tdb->exts[k];    /* add the extension train */
      d = proj->mem;            /* and filter the collected trains */
      e = tna_seqfiltx(proj->trains, tdb->trains, n+1, proj->ivals,
                       fabs(ccn->width), ccn->supp, ccn->used, &d);
      if (e < n+1) continue;    /* if one becomes empty, skip ext. */
      s = seqsupp(proj->trains, n+1, ccn->width, ccn->heap);
      if (s < ccn->supp)        /* compute the sequence support, */
        continue;               /* skip infrequent sequences, and */
      if (s > m) m = s;         /* find maximum extension support */
      i = trn_item(tdb->trains[n]); /* get current extension item */
      r = isr_addnc(ccn->report, i, s);
      if (r < 0) break;         /* add current item to the reporter */
      if ((u = ccn->used[i]) < 0)  /* if the train has not been used, */
        ccn->used[i] = n;          /* note current as first use index */
      /* filter/reduce the extension trains (reduce times/points) */
      e = (ccn->target & (CCN_CLOSED|CCN_MAXIMAL))
        ? tna_seqred (proj->exts, tdb->exts, x, proj->ivals,
                      fabs(ccn->width), ccn->supp, -1, &d)
        : tna_seqredx(proj->exts, tdb->exts, x, proj->ivals,
                      fabs(ccn->width), ccn->supp,
                      ccn->used, proj->trains, &d);
      proj->size = (size_t)((char*)d -(char*)proj->mem);
      r = rec_seqfilter(ccn, proj, n+1, e, s);
      isr_remove(ccn->report,1);/* recursively find freq. item seqs. */
      ccn->used[i] = u;         /* remove current item from reporter */
      if (r < 0) break;         /* and restore its first use index */
    }                           /* if an error occurred, abort */
    free(proj);                 /* deallocate the projection memory */
    if (r < 0) return r;        /* if an error occurred, abort */
  }
  if (!(ccn->target & (CCN_CLOSED|CCN_MAXIMAL))
  ||  seqclomax(ccn, tdb->trains, n, tdb->exts, x, supp, m))
    r = isr_report(ccn->report);/* report the current item sequence */
  return r;                     /* return the error status */
}  /* rec_seqfilter() */

/*--------------------------------------------------------------------*/

int ccn_seqfilter (COCONAD *ccn)
{                               /* --- closed cont. neuron ass. dect. */
  int     r = 0;                /* result of recursion/error status */
  ITEM    i, k, n;              /* loop variables, number of items */
  TID     s;                    /* size/length of a train */
  TRAIN   *t, **trains;         /* to traverse the trains */
  TRAINDB *tdb;                 /* train database for recursion */
  IVAL    *ival;                /* to initialize the intervals */

  assert(ccn);                  /* check the function arguments */
  if (ccn->supp > (SUPP)tns_maxlen(ccn->trnset))
    return 0;                   /* check the maximum train length */
  k = tns_cnt(ccn->trnset);     /* get the number of items and */
  if (k <= 0) return 0;         /* check for at least one item */
  if (ccn->target & CCN_PERMUT){/* if to find item permutations */
    ccn->used = NULL;           /* no first used array is needed and */
    n = k; }                    /* maximum size is number of items */
  else {                        /* if to find item sequences */
    ccn->used = (ITEM*)malloc((size_t)k *sizeof(ITEM));
    if (!ccn->used) return -1;  /* create array of first use indices */
    for (i = 0; i < k; i++) ccn->used[i] = -1;
    n = tns_maxseq(ccn->trnset, fabs(ccn->width)) +1;
    if (n < k) n = k;           /* get maximum pattern length and */
  }                             /* ensure at least number of items */
  tdb = (TRAINDB*)malloc(sizeof(TRAINDB)  +sizeof(IVAL)
                         +(size_t)(1+k+n) *sizeof(TRAIN*)
                         +(size_t) n      *sizeof(double*));
  if (!tdb) { if (ccn->used) { free(ccn->used); } return -1; }
  tdb->trains = trains  = (TRAIN**)(tdb->ivals+2);
  tdb->exts   = trains += 1;    /* get and organize memory, that is, */
  tdb->mem    = NULL;           /* set up the different train arrays */
  tdb->size   = tna_dstsize(k, tns_extent(ccn->trnset));
  ccn->excl   = trains +k;      /* note the heap for support comp. */
  ccn->heap   = (const double**)(ccn->excl +n);
  ival = tdb->ivals;   ival->beg = -INFINITY; ival->end = INFINITY;
  ival = tdb->ivals+1; ival->beg =  INFINITY; ival->end = INFINITY;
  for (ccn->max = 0, i = n = 0; i < k; i++) {
    t = tns_train(ccn->trnset, i);    /* traverse the trains, but */
    s = trn_len(t);             /* skip trains that are too short */
    if ((SUPP)s < ccn->supp) continue;
    if (s > ccn->max) ccn->max = s;
    trains[n++] = t;            /* find the maximum train size and */
  }                             /* collect sufficiently long trains */
  if (n > 0) {                  /* if there are frequent items */
    r = (ccn->target & CCN_SEQUENCE)
      ? rec_seqfilter(ccn, tdb, 0, n, SUPP_MAX)
      : rec_pmtfilter(ccn, tdb, 0, n, SUPP_MAX);
  }                             /* recursively find freq. sequences */
  free(tdb);                    /* deallocate the work memory */
  if (ccn->used) free(ccn->used);
  return (r < 0) ? r : 0;       /* return the error status */
}  /* ccn_seqfilter() */

/*----------------------------------------------------------------------
  CoCoNAD (closed/maximal pattern filtering with repository)
----------------------------------------------------------------------*/

static int rec_pmtrepo (COCONAD *ccn, TRAINDB *tdb, ITEM n, ITEM x)
{                               /* --- recursive part of coconad */
  int     r = 0;                /* result of functions/error status */
  ITEM    k;                    /* loop variable for extensions */
  ITEM    e;                    /* number of extension items */
  size_t  z;                    /* size of conditional database */
  SUPP    s, m = 0;             /* (maximum) support (of extension) */
  TRAINDB *proj = NULL;         /* projected/filtered train database */
  void    *d;                   /* pointer to free destination memory */

  assert(ccn && tdb             /* check the function arguments */
  &&    (n >= 0) && (x >= 0));
  #ifdef CCN_ABORT              /* if to check for interrupt */
  if (sig_aborted()) return -1; /* if execution was aborted, */
  #endif                        /* abort the recursion */
  if ((x > 0)                   /* if there are extension items */
  &&  isr_xable(ccn->report,1)){/* and another item can be added */
    z = (size_t)n+2 +(size_t)x; /* compute max. number of trains */
    z += z & 1;                 /* (even number for alignment) */
    proj = (TRAINDB*)malloc(sizeof(TRAINDB)
                          +(size_t)ccn->max *sizeof(IVAL)
                          +z *sizeof(TRAIN*) +tdb->size);
    if (!proj) return -1;       /* allocate memory for projection */
    proj->trains = (TRAIN**)(proj->ivals +ccn->max +1);
    proj->exts   = proj->trains +n+2;
    proj->mem    = (void*)(proj->trains +z);
    for (k = 0; k < x; k++) {   /* traverse possible extensions */
      s = (SUPP)trn_seqisect(proj->ivals, tdb->ivals,
                             tdb->exts[k], fabs(ccn->width));
      if (s < ccn->supp)        /* intersect the interval list and */
        continue;               /* check the number of used spikes */
      tdb->trains[n] = tdb->exts[k];    /* add the extension train */
      d = proj->mem;            /* and filter the collected trains */
      e = tna_seqfilt(proj->trains, tdb->trains, n+1, proj->ivals,
                      fabs(ccn->width), ccn->supp, &d);
      if (e < n+1) continue;    /* if one becomes empty, skip ext. */
      s = seqsupp(proj->trains, n+1, ccn->width, ccn->heap);
      if (s < ccn->supp)        /* compute the sequence support, */
        continue;               /* skip infrequent sequences, and */
      if (s > m) m = s;         /* find maximum extension support */
      r = isr_addnc(ccn->report, trn_item(tdb->trains[n]), s);
      if (r < 0) break;         /* add current item to the reporter */
      if (ccn->mode & CCN_REPOCHAIN) { /* if tree chain filter */
        r = cmx_add(ccn->cmx, trn_item(tdb->trains[n]), s);
        if (r < 0) { isr_remove(ccn->report, 1); break; }
      }                         /* add item to closed/maximal filter */
      tdb->exts[k] = NULL;      /* (temporarily) remove its train */
      /* filter/reduce the extension trains (reduce times/points) */
      e = tna_seqred(proj->exts, tdb->exts, x, proj->ivals,
                     fabs(ccn->width), ccn->supp,
                     ccn->target & (CCN_CLOSED|CCN_MAXIMAL), &d);
      proj->size = (size_t)((char*)d -(char*)proj->mem);
      r = rec_pmtrepo(ccn, proj, n+1, e);  /* recursively find seqs. */
      isr_remove(ccn->report,1);/* remove current item from reporter */
      if (ccn->mode & CCN_REPOCHAIN)
        cmx_remove(ccn->cmx, 1);/* remove item also from c/m filter */
      tdb->exts[k] = tdb->trains[n];
      if (r < 0) break;         /* restore the current item's train */
    }                           /* if an error occurred, abort */
    free(proj);                 /* deallocate the projection memory */
    if (r < 0) return r;        /* if an error occurred, abort */
  }
  return (ccn->repfn) ? ccn->repfn(ccn, m) : isr_report(ccn->report);
}  /* rec_pmtrepo() */          /* report the current permutation */

/*--------------------------------------------------------------------*/

static int rec_seqrepo (COCONAD *ccn, TRAINDB *tdb, ITEM n, ITEM x)
{                               /* --- recursive part of coconad */
  int     r = 0;                /* result of functions/error status */
  ITEM    k;                    /* loop variable for extensions */
  ITEM    i, u;                 /* extension item and its first use */
  ITEM    e;                    /* number of extension items */
  size_t  z;                    /* size of conditional database */
  SUPP    s, m = 0;             /* (maximum) support (of extension) */
  TRAINDB *proj = NULL;         /* projected/filtered train database */
  void    *d;                   /* pointer to free destination memory */

  assert(ccn && tdb             /* check the function arguments */
  &&    (n >= 0) && (x >= 0));
  #ifdef CCN_ABORT              /* if to check for interrupt */
  if (sig_aborted()) return -1; /* if execution was aborted, */
  #endif                        /* abort the recursion */
  if ((x > 0)                   /* if there are extension items */
  &&  isr_xable(ccn->report,1)){/* and another item can be added */
    z = (size_t)n+2 +(size_t)x; /* compute max. number of trains */
    z += z & 1;                 /* (even number for alignment) */
    proj = (TRAINDB*)malloc(sizeof(TRAINDB)
                          +(size_t)ccn->max *sizeof(IVAL)
                          +z *sizeof(TRAIN*) +tdb->size);
    if (!proj) return -1;       /* allocate memory for projection */
    proj->trains = (TRAIN**)(proj->ivals +ccn->max +1);
    proj->exts   = proj->trains +n+2;
    proj->mem    = (void*)(proj->trains +z);
    for (k = 0; k < x; k++) {   /* traverse possible extensions */
      s = (SUPP)trn_seqisect(proj->ivals, tdb->ivals,
                             tdb->exts[k], fabs(ccn->width));
      if (s < ccn->supp)        /* intersect the interval list and */
        continue;               /* check the number of used spikes */
      tdb->trains[n] = tdb->exts[k];    /* add the extension train */
      d = proj->mem;            /* and filter the collected trains */
      e = tna_seqfiltx(proj->trains, tdb->trains, n+1, proj->ivals,
                       fabs(ccn->width), ccn->supp, ccn->used, &d);
      if (e < n+1) continue;    /* if one becomes empty, skip ext. */
      s = seqsupp(proj->trains, n+1, ccn->width, ccn->heap);
      if (s < ccn->supp)        /* compute the sequence support, */
        continue;               /* skip infrequent sequences, and */
      if (s > m) m = s;         /* find maximum extension support */
      i = trn_item(tdb->trains[n]); /* get current extension item */
      r = isr_addnc(ccn->report, i, s);
      if (r < 0) break;         /* add current item to the reporter */
      if (ccn->mode & CCN_REPOCHAIN) {  /* if tree chain filter */
        r = cmx_add(ccn->cmx, trn_item(tdb->trains[n]), s);
        if (r < 0) { isr_remove(ccn->report, 1); break; }
      }                         /* add item to closed/maximal filter */
      if ((u = ccn->used[i]) < 0) /* if the train has not been used, */
        ccn->used[i] = n;         /* note current as first use index */
      /* filter/reduce the extension trains (reduce times/points) */
      e = tna_seqredx(proj->exts, tdb->exts, x, proj->ivals,
                      fabs(ccn->width), ccn->supp,
                      ccn->used, proj->trains, &d);
      proj->size = (size_t)((char*)d -(char*)proj->mem);
      r = rec_seqrepo(ccn, proj, n+1, e);  /* recursively find seqs. */
      isr_remove(ccn->report,1);/* remove current item from reporter */
      if (ccn->mode & CCN_REPOCHAIN)
        cmx_remove(ccn->cmx, 1);/* remove item also from c/m filter */
      ccn->used[i] = u;         /* restore the first use index */
      if (r < 0) break;         /* if an error occurred, abort */
    }
    free(proj);                 /* deallocate the projection memory */
    if (r < 0) return r;        /* if an error occurred, abort */
  }
  return (ccn->repfn) ? ccn->repfn(ccn, m) : isr_report(ccn->report);
}  /* rec_seqrepo() */          /* report the current sequence */

/*--------------------------------------------------------------------*/

int ccn_seqrepo (COCONAD *ccn)
{                               /* --- closed cont. neuron ass. dect. */
  int     r;                    /* result of recursion/error status */
  int     trg;                  /* target pattern type for filter */
  ITEM    i, k, n;              /* loop variables, number of items */
  TID     s;                    /* size/length of a train */
  TRAIN   *t, **trains;         /* to traverse the trains */
  TRAINDB *tdb;                 /* train database for recursion */
  IVAL    *ival;                /* to initialize the intervals */
  void    *p = NULL;            /* typeless repository */

  assert(ccn);                  /* check the function arguments */
  if (!(ccn->target & (CCN_CLOSED|CCN_MAXIMAL)))
       ccn->mode = (ccn->mode & ~CCN_REPOTYPE);
  else if (ccn->mode & CCN_REPOLIST) /* make search mode consistent */
       ccn->mode = (ccn->mode & ~CCN_REPOTYPE) | CCN_REPOLIST;
  else if (ccn->mode & CCN_REPOTREE) /* (only one repo. flag set) */
       ccn->mode = (ccn->mode & ~CCN_REPOTYPE) | CCN_REPOTREE;
  else ccn->mode = (ccn->mode & ~CCN_REPOTYPE) | CCN_REPOCHAIN;
  if (ccn->supp > (SUPP)tns_maxlen(ccn->trnset))
    return 0;                   /* check the maximum train length */
  k = tns_cnt(ccn->trnset);     /* get the number of items */
  if (k <= 0) return 0;         /* and the maximum pattern length */
  if (ccn->target & CCN_PERMUT){/* if to find item permutations, */
    ccn->used = NULL;           /* no first used array is needed and */
    n = k; }                    /* maximum size is number of items */
  else {                        /* if to find item sequences */
    ccn->used = (ITEM*)malloc((size_t)k *sizeof(ITEM));
    if (!ccn->used) return -1;  /* create array of first use indices */
    for (i = 0; i < k; i++) ccn->used[i] = -1;
    n = tns_maxseq(ccn->trnset, fabs(ccn->width)) +1;
    if (n < k) n = k;           /* get maximum pattern length and */
  }                             /* ensure at least number of items */
  ccn->cmf = NULL;
  ccn->cmx = NULL;              /* clear all repository variables */
  ccn->repfn = (REPORTFN*)0;    /* and the pattern report function */
  if (ccn->target & (CCN_CLOSED|CCN_MAXIMAL)) {
    if (ccn->mode & CCN_REPOLIST) { /* if repository is list/array */
      ccn->cmf = p = cmf_create((size_t)(1024*1024));
      ccn->repfn = report2cmf;} /* create it and get update function */
    else {                      /* if repo. is a (chain of) tree(s) */
      trg = ((ccn->target & CCN_PERMUT) ? CMX_PERMUT : CMX_SEQUENCE)
          | ((ccn->target & CCN_CLOSED) ? CMX_CLOSED : CMX_MAXIMAL);
      ccn->cmx = p = cmx_create(k, n, trg, -1);
      ccn->repfn = report2cmx;  /* create it and get update function */
    }                           /* (set up the filter repository) */
    if (!p) { if (ccn->used) { free(ccn->used); } return -1; }
  }                             /* check for creation success */
  tdb = (TRAINDB*)malloc(sizeof(TRAINDB) +sizeof(IVAL)
                         +(size_t)(1+k)  *sizeof(TRAIN*)
                         +(size_t) n     *sizeof(double*));
  if (!tdb) { ccnclean(ccn); return -1; }
  tdb->trains = trains = (TRAIN**)(tdb->ivals+2);
  tdb->exts   = trains += 1;    /* get and organize memory, that is, */
  tdb->mem    = NULL;           /* set up the different train arrays */
  tdb->size   = tna_dstsize(k, tns_extent(ccn->trnset));
  ccn->heap   = (const double**)(trains +k);
  ival = tdb->ivals;   ival->beg = -INFINITY; ival->end = INFINITY;
  ival = tdb->ivals+1; ival->beg =  INFINITY; ival->end = INFINITY;
  for (ccn->max = 0, i = n = 0; i < k; i++) {
    t = tns_train(ccn->trnset, i);    /* traverse the trains, but */
    s = trn_len(t);             /* skip trains that are too short */
    if ((SUPP)s < ccn->supp) continue;
    if (s > ccn->max) ccn->max = s;
    trains[n++] = t;            /* find the maximum train size and */
  }                             /* collect sufficiently long trains */
  r = (n <= 0) ? 0 : (ccn->target & CCN_SEQUENCE)
    ? rec_seqrepo(ccn, tdb, 0, n)   /* recursively find */
    : rec_pmtrepo(ccn, tdb, 0, n);  /* frequent sequences */
  free(tdb);                    /* deallocate the work memory */
  if (r >= 0) {                 /* if to report from a repository */
    if      (ccn->cmx) r = cmx_report(ccn->cmx, ccn->report);
    else if (ccn->cmf) r = (ccn->target & CCN_MAXIMAL)
                         ? cmf_maximal(ccn->cmf, ccn->report)
                         : cmf_closed (ccn->cmf, ccn->report);
  }                             /* report closed/maximal sequences */
  ccnclean(ccn);                /* clean up possible repositories */
  return (r < 0) ? r : 0;       /* return the error status */
}  /* ccn_seqrepo() */

#endif  /* #ifndef CCN_OVERLAP */
/*----------------------------------------------------------------------
  CoCoNAD (overlap based support, interval list intersection)
----------------------------------------------------------------------*/
#ifdef CCN_OVERLAP              /* if influence map overlap support */

static int rec_ovlrepo (COCONAD *ccn, IVLIST **ivls, ITEM k, size_t x)
{                               /* --- recursive part of coconad */
  int    r = 0;                 /* result of functions/error status */
  ITEM   i, m;                  /* loop variables for items */
  SUPP   pex;                   /* minimum support for perfect exts. */
  IVLIST *l, *d;                /* to traverse the interval lists */
  IVLIST **proj = NULL;         /* interval lists of proj. database */
  size_t y, z;                  /* intervals in intersection */

  assert(ccn && ivls && (k >= 0)); /* check the function arguments */
  #ifdef CCN_ABORT              /* if to check for interrupt */
  if (sig_aborted()) return -1; /* if execution was aborted, */
  #endif                        /* abort the recursion */
  if ((k > 1)                   /* if there is more than one item */
  &&  isr_xable(ccn->report,2)){/* and another item can be added */
    z = x;                      /* find maximally needed intervals */
    for (i = k; --i >= 0; ) {   /* traverse the items/interval lists */
      y = (size_t)((IVAL*)ivls[i+1] -ivls[i]->ivals) -1;
      x -= y; y = x +(y-1)*(size_t)i;
      if (y > z) z = y;         /* get number of intervals in list */
    }                           /* and maximal size of intersections */
    proj = (IVLIST**)malloc((size_t)k *sizeof(IVLIST*)
                           +(size_t)k *sizeof(IVLIST)
                           +        z *sizeof(IVAL));
    if (!proj) return -1;       /* allocate list and element arrays */
  }                             /* (memory for conditional databases) */
  while (--k >= 0) {            /* traverse the extension items */
    l = ivls[k];                /* get interval list for current item */
    r = isr_add(ccn->report, l->item, l->supp);
    if (r <  0) break;          /* add current item to the reporter */
    if (r <= 0) continue;       /* check if item needs processing */
    if (proj && (k > 0)) {      /* if another item can be added */
      pex = (ccn->mode & CCN_PERFECT) ? l->supp : SUPP_MAX;
      proj[m = 0] = d = (IVLIST*)(proj +k+1); x = 0;
      for (i = 0; i < k; i++) { /* intersect with preceding lists */
        z = isect(d, ivls[i], l, ccn->width);
        if (d->supp >= pex) {   /* collect perfect extensions */
          isr_addpex(ccn->report, d->item); continue; }
        if (d->supp >= ccn->supp) { /* collect frequent extensions */
          proj[++m] = d = (IVLIST*)(d->ivals +z+1); x += z; }
      }                         /* switch to the next output list */
      if (m > 0) {              /* if the projection is not empty */
        r = rec_ovlrepo(ccn, proj, m, x);
        if (r < 0) break;       /* recursively find freq. item sets */
      }                         /* in the created projection */
    }
    r = isr_report(ccn->report);/* report the current item set */
    if (r < 0) break;           /* and check for an error */
    isr_remove(ccn->report, 1); /* remove the current item */
  }                             /* from the item set reporter */
  if (proj) free(proj);         /* delete the list and element arrays */
  return r;                     /* return the error status */
}  /* rec_ovlrepo() */

/*--------------------------------------------------------------------*/

int ccn_ovlrepo (COCONAD *ccn)
{                               /* --- closed cont. neuron ass. dect. */
  int    r = 0;                 /* result of recursion/error status */
  ITEM   i, k, n;               /* loop variable, number of items */
  IVLIST *l, **ivls;            /* to traverse the interval lists */
  size_t x, z;                  /* (sum of) interval list sizes */

  assert(ccn);                  /* check the function arguments */
  if (ccn->supp > (SUPP)tns_maxlen(ccn->trnset))
    return 0;                   /* check the maximum train length */
  k = tns_cnt(ccn->trnset);     /* get the number of items and */
  if (k <= 0) return 0;         /* check for at least one item */
  n = k +2 -(k & 1);            /* make number of pointers even */
  ivls = (IVLIST**)malloc((size_t)n *sizeof(IVLIST*)
                         +(size_t)k *sizeof(IVLIST)
                         +tns_extent(ccn->trnset) *sizeof(IVAL));
  if (!ivls) return -1;         /* allocate the work memory and */
  ivls[0] = l = (IVLIST*)(ivls +n);  /* get first interval list */
  for (i = n = 0, x = 0; i < k; i++) {   /* traverse the trains */
    z = trn2ivals(tns_train(ccn->trnset, i), ccn->width, l);
    if (l->supp < ccn->supp) continue;
    ivls[++n] = l = (IVLIST*)(l->ivals +z+1); x += z;
  }                             /* turn trains into interval lists */
  r = (n <= 0) ? 0              /* recursively find freq. item sets */
    : rec_ovlrepo(ccn, ivls, n, x);
  free(ivls);                   /* delete the allocated arrays */
  return (r < 0) ? r : 0;       /* return the error status */
}  /* ccn_ovlrepo() */

#endif  /* #ifdef CCN_OVERLAP */
/*----------------------------------------------------------------------
  CoCoNAD (cover similarity support)
----------------------------------------------------------------------*/
#ifdef CCN_OVERLAP              /* if influence map overlap support */

static int rec_simrepo (COCONAD *ccn, XIVLIST **ivls, ITEM k, size_t x)
{                               /* --- recursive part of coconad */
  int     r = 0;                /* result of functions/error status */
  ITEM    i, m;                 /* loop variables */
  XIVLIST *l, *d;               /* to traverse the interval lists */
  XIVLIST **proj = NULL;        /* interval lists of proj. database */
  size_t  y, z;                 /* intervals in intersection */

  assert(ccn && ivls && (k >= 0)); /* check the function arguments */
  #ifdef CCN_ABORT              /* if to check for interrupt */
  if (sig_aborted()) return -1; /* if execution was aborted, */
  #endif                        /* abort the recursion */
  if ((k > 1)                   /* if there is more than one item */
  &&  isr_xable(ccn->report,2)){/* and another item can be added */
    z = x;                      /* find maximally needed intervals */
    for (i = k; --i >= 0; ) {   /* traverse the items/interval lists */
      y = (size_t)((IVAL*)ivls[i+1] -ivls[i]->ivals) -1;
      x -= y; y = x +(y-1)*(size_t)i;
      if (y > z) z = y;         /* get number of intervals in list */
    }                           /* and maximal size of intersections */
    proj = (XIVLIST**)malloc((size_t)k *sizeof(XIVLIST*)
                            +(size_t)k *sizeof(XIVLIST)
                            +        z *sizeof(IVAL));
    if (!proj) return -1;       /* allocate list and element arrays */
  }                             /* (memory for conditional databases) */
  while (--k >= 0) {            /* traverse the extension items */
    l = ivls[k];                /* get interval list for current item */
    r = (ccn->mode & CCN_SIMSUPP)
      ? isr_addwgt(ccn->report, l->item, l->sim, l->all)
      : isr_addwgt(ccn->report, l->item, l->all, l->sim);
    if (r <  0) break;          /* add current item to the reporter */
    if (r <= 0) continue;       /* check if item needs processing */
    if (proj && (k > 0)) {      /* if another item can be added */
      proj[m = 0] = d = (XIVLIST*)(proj +k+1); x = 0;
      for (i = 0; i < k; i++) { /* intersect with preceding lists */
        xisect(d, ivls[i], l, ccn->width);
        if (d->all < ccn->supp) continue;
        z = xmerge(d, ivls[i], l, ccn->width);
        d->sim = ccn->simfn(d->all, d->one, ccn->total, ccn->simps);
        if (d->sim < ccn->sim)  /* compute the similarity measure and */
          continue;             /* check it against given minimum */
        proj[++m] = d = (XIVLIST*)(d->ivals +z+1); x += z;
      }                         /* switch to the next output list */
      if (m > 0) {              /* if the projection is not empty */
        r = rec_simrepo(ccn, proj, m, x);
        if (r < 0) break;       /* recursively find freq. item sets */
      }                         /* in the created projection */
    }
    r = isr_report(ccn->report);/* report the current item set */
    if (r < 0) break;           /* and check for an error */
    isr_remove(ccn->report, 1); /* remove the current item */
  }                             /* from the item set reporter */
  if (proj) free(proj);         /* delete the list and element arrays */
  return r;                     /* return the error status */
}  /* rec_simrepo() */

/*--------------------------------------------------------------------*/

int ccn_simrepo (COCONAD *ccn, int simcov, double *simps,
                 double sim, double beg, double end)
{                               /* --- closed cont. neuron ass. dect. */
  int     r = 0;                /* result of recursion/error status */
  ITEM    i, k, n;              /* loop variable, number of items */
  XIVLIST *l, **ivls;           /* to traverse the interval lists */
  size_t  x, z;                 /* (sum of) interval list sizes */

  assert(ccn);                  /* check the function arguments */
  if (ccn->supp > (SUPP)tns_maxlen(ccn->trnset))
    return 0;                   /* check the maximum train length */
  k = tns_cnt(ccn->trnset);     /* get the number of items and */
  if (k <= 0) return 0;         /* check for at least one item */
  if ((beg <= -INFINITY) || (beg >= INFINITY))
    beg = floor(tns_min(ccn->trnset, 0)); /* get start of event span */
  if ((end <= -INFINITY) || (end >= INFINITY))
    end = ceil (tns_max(ccn->trnset, 0)); /* get end   of event span */
  ccn->total = (ccn->width > 0) ? (end-beg) /ccn->width : INFINITY;
  if (simcov < CCN_GENERIC)     /* if specific similarity measure, */
    ccn->simfn = simfns[simcov];/* store the corresponding function */
  else {                        /* if generic similarity measure */
    ccn->simps[0] = simps[0];
    ccn->simps[4] = simps[4];
    ccn->simps[1] =          -simps[1];
    ccn->simps[5] =          -simps[5];
    ccn->simps[2] = simps[2] +simps[1];
    ccn->simps[6] = simps[6] +simps[5];
    ccn->simps[3] = simps[3];
    ccn->simps[7] = simps[7];
    if      ((fabs(ccn->simps[3]) > DBL_MIN)
    ||       (fabs(ccn->simps[7]) > DBL_MIN))
         ccn->simfn = generic_0;/* use all parameters */
    else if  (fabs(ccn->simps[2]) > DBL_MIN)
         ccn->simfn = generic_1;/* do not use geometric mean of s & z */
    else if  (fabs(ccn->simps[1]) > DBL_MIN)
         ccn->simfn = generic_2;/* do not use zero-zero counter */
    else ccn->simfn = generic_3;/* use normalized support */
  }                             /* (set similarity measure function) */
  ccn->sim = (sim > 0) ? sim : 0;   /* note the similarity threshold */
  n    = k +2 -(k & 1);         /* make number of pointers even */
  ivls = (XIVLIST**)malloc((size_t)n *sizeof(XIVLIST*)
                          +(size_t)k *sizeof(XIVLIST)
                          +2*tns_extent(ccn->trnset) *sizeof(IVAL));
  if (!ivls) return -1;         /* allocate the work memory and */
  ivls[0] = l = (XIVLIST*)(ivls +n); /* get first interval list */
  for (i = n = 0, x = 0; i < k; i++) {   /* traverse the trains */
    z = trn2xivls(tns_train(ccn->trnset, i), ccn->width, beg, end, l);
    if (l->all < ccn->supp) continue;
    l->sim = ccn->simfn(l->all, l->one, ccn->total, ccn->simps);
    if (l->sim < ccn->sim)  continue;
    ivls[++n] = l = (XIVLIST*)(l->ivals +z+1); x += z;
  }                             /* turn trains into interval lists */
  r = (n <= 0) ? 0              /* recursively find freq. item sets */
    : rec_simrepo(ccn, ivls, n, x);
  free(ivls);                   /* delete the allocated arrays */
  return (r < 0) ? r : 0;       /* return the error status */
}  /* ccn_simrepo() */

#endif
/*----------------------------------------------------------------------
  CoCoNAD (generic)
----------------------------------------------------------------------*/

COCONAD* ccn_create (int target, double width, double smin,
#ifdef CCN_OVERLAP              /* if influence map overlap support */
                     double mix,
#endif                          /* add. support mixing parameter */
                     ITEM zmin, ITEM zmax, int algo, int mode)
{                               /* --- create a CoCoNAD miner */
  COCONAD *ccn;                 /* created coconad miner */
  int     t;                    /* target pattern subtype */

  /* --- make parameters consistent --- */
  if      (target & CCN_MAXIMAL) t = ISR_MAXIMAL;
  else if (target & CCN_CLOSED)  t = ISR_CLOSED;
  else                           t = ISR_FREQUENT;
  target = (target & CCN_TYPEMASK) | t;
  #ifdef CCN_OVERLAP            /* if influence map overlap support */
  target = (target & CCN_SUBMASK) | CCN_ITEMSET;
  if ((mix > 0) && (algo == CCN_ISECT))
    algo = CCN_REPO;            /* only item set mining is possible */
  #endif                        /* and no list intersection with mix */
  if (algo == CCN_AUTO) algo = CCN_REPO;
  if ((target & (CCN_CLOSED|CCN_MAXIMAL))
  && (algo != CCN_BASIC) && (algo != CCN_FILTER))
    mode &= ~CCN_REORDER;       /* reordering is not always possible */

  /* --- create a coconad miner --- */
  ccn = (COCONAD*)malloc(sizeof(COCONAD));
  if (!ccn) return NULL;        /* create an coconad miner */
  ccn->target = target;         /* and store all parameters */
  ccn->width  = width;
  ccn->supp   = ((SUPP)smin > 0) ? (SUPP)smin : 1;
  ccn->zmin   = zmin;
  ccn->zmax   = zmax;
  ccn->xmax   = zmax;
  ccn->algo   = algo;
  ccn->mode   = mode;
  ccn->trnset = NULL;
  ccn->report = NULL;
  ccn->limit  = 0;
  ccn->max    = 0;
  ccn->hsize  = 0;
  ccn->heap   = NULL;
  ccn->pexs   = NULL;
  ccn->excl   = NULL;
  ccn->used   = NULL;
  ccn->cmf    = NULL;
  #ifndef CCN_OVERLAP           /* if standard (MIS) support */
  ccn->cmx    = NULL;           /* closed/maximal sequence tree chain */
  #else                         /* if influence map overlap support */
  ccn->total  = 0;              /* maximal theoret. possible support */
  ccn->simfn  = (SIMFN*)0;      /* similarity measure function */
  ccn->sim    = 0;              /* minimum similarity measure value */
  ccn->mix    = mix;            /* mixing coefficient for standard */
  #endif                        /* and influence map overlap support */
  return ccn;                   /* return the created coconad miner */
}  /* ccn_create() */

/*--------------------------------------------------------------------*/

void ccn_delete (COCONAD *ccn, int deldar)
{                               /* --- delete an eclat miner */
  if (deldar) {                 /* if to delete data and reporter */
    if (ccn->report) isr_delete(ccn->report, 0);
    if (ccn->trnset) tns_delete(ccn->trnset, 1);
  }                             /* delete if existing */
  free(ccn);                    /* delete the base structure */
}  /* ccn_delete() */

/*--------------------------------------------------------------------*/

int ccn_data (COCONAD *ccn, TRAINSET *trnset, int mode)
{                               /* --- prepare data for CoCoNAD */
  #ifndef QUIET                 /* if to print messages */
  ITEM    m;                    /* number of items/trains */
  size_t  n;                    /* total number of points */
  clock_t t;                    /* timer for measurements */
  #endif                        /* (only needed for messages) */

  assert(ccn && trnset);        /* check the function arguments */
  ccn->trnset = trnset;         /* note the train set */
  CLOCK(t);                     /* start timer, print log message */
  XMSG(stderr, "sorting events in trains ... ");
  tns_sort(ccn->trnset, +1);    /* sort the trains by points */
  #ifdef CCN_OVERLAP            /* if influence map overlap support */
  if (((ccn->algo == CCN_REPO)  /* if to filter with a repository */
  ||   (ccn->algo == CCN_ISECT))/* or to use intersection algorithm */
  &&  !(mode & CCN_NORECODE))   /* and recoding is permitted, */
    tns_recode(ccn->trnset,-1); /* sort items by descending freq. */
  #else                         /* if standard (MIS) support */
  if ((ccn->algo == CCN_REPO)   /* if to filter with a repository */
  &&  !(mode & CCN_NORECODE))   /* and recoding is permitted, */
    tns_recode(ccn->trnset,-1); /* sort items by descending freq. */
  #endif
  #ifndef QUIET                 /* if to print messages */
  m = tns_cnt(ccn->trnset);     /* get the number of trains */
  n = tns_extent(ccn->trnset);  /* and the number of points */
  XMSG(stderr, "[%"ITEM_FMT" train(s), %"SIZE_FMT" event(s)]", m, n);
  XMSG(stderr, " done [%.2fs].\n", SEC_SINCE(t));
  #endif                        /* print a log message */
  return 0;                     /* return 'ok' */
}  /* ccn_data() */

/*--------------------------------------------------------------------*/

extern int ccn_report (COCONAD *ccn, ISREPORT *report)
{                               /* --- prepare reporter for CoCoNAD */
  int mrep;                     /* mode for item set reporter */

  assert(ccn && report);        /* check the function arguments */
  ccn->report = report;         /* note the item set reporter */
  mrep = (ccn->target & CCN_SEQUENCE) ? ISR_SEQUENCE|ISR_NOFILTER
       : ((ccn->algo == CCN_BASIC) || (ccn->algo == CCN_FILTER))
       ? ISR_NOFILTER : 0;      /* build the reporting mode */
  isr_setsupp(report, (RSUPP)ccn->supp, RSUPP_MAX);
  isr_setsize(report, ccn->zmin, ccn->zmax);
  isr_setsmt(report, (RSUPP)1); /* set support/weight of empty set */
  isr_setwgt(report, 1.0);      /* configure the item set reporter */
  return (isr_settarg(report, ccn->target, mrep, -1)) ? E_NOMEM : 0;
}  /* ccn_report() */

/*--------------------------------------------------------------------*/

int ccn_mine (COCONAD *ccn)
{                               /* --- main CoCoNAD function */
  int     r;                    /* result of function call */
  #ifndef QUIET                 /* if to print messages */
  clock_t t;                    /* timer for measurements */
  #endif                        /* (only needed for messages) */

  assert(ccn);                  /* check the function arguments */
  CLOCK(t);                     /* start timer, print log message */
  XMSG(stderr, "writing %s ... ", isr_name(ccn->report));
  #ifdef CCN_OVERLAP            /* if influence map overlap support */
  if      (ccn->algo == CCN_ISECT)
    r = ccn_ovlrepo(ccn);       /* intersection algorithm */
  else if (ccn->algo == CCN_FILTER)
    r = ccn_filter (ccn);       /* point and train filtering */
  else if (ccn->algo == CCN_REPO)
    r = ccn_cmrepo (ccn);       /* closed/maximal with a repository */
  else
    r = ccn_basic  (ccn);       /* basic algorithm (fallback) */
  #else                         /* if integer (MIS) support type */
  if      (ccn->algo == CCN_FILTER) /* point and train filtering */
    r = (ccn->target & (CCN_PERMUT|CCN_SEQUENCE))
      ? ccn_seqfilter(ccn) : ccn_filter(ccn);
  else if (ccn->algo == CCN_REPO)   /* closed/maximal repository */
    r = (ccn->target & (CCN_PERMUT|CCN_SEQUENCE))
      ? ccn_seqrepo  (ccn) : ccn_cmrepo(ccn);
  else                              /* basic algorithm (fallback) */
    r = (ccn->target & (CCN_PERMUT|CCN_SEQUENCE))
      ? ccn_seqbasic (ccn) : ccn_basic (ccn);
  #endif                        /* search for frequent patterns */
  if (r < 0) return E_NOMEM;    /* check for a search error */
  XMSG(stderr, "[%"ISCNT_FMT" pattern(s)]",
       ISCOUT(isr_repcnt(ccn->report)));
  XMSG(stderr, " done [%.2fs].\n", SEC_SINCE(t));
  return 0;                     /* return 'ok' */
}  /* ccn_mine() */

/*--------------------------------------------------------------------*/
#ifdef CCN_OVERLAP              /* if influence map overlap support */

int ccn_jim (COCONAD *ccn, int simcov, double *simps, double sim,
             double beg, double end)
{                               /* --- main CoCoNAD function */
  int     r;                    /* result of function call */
  #ifndef QUIET                 /* if to print messages */
  clock_t t;                    /* timer for measurements */
  #endif                        /* (only needed for messages) */

  assert(ccn);                  /* check the function arguments */
  sim /= 100.0;                 /* scale the similarity threshold */
  #ifdef CCN_OVERLAP            /* if influence map overlap support */
  isr_setsupp(ccn->report, (RSUPP)((ccn->mode & CCN_SIMSUPP)
                                   ? sim : ccn->supp), RSUPP_MAX);
  #endif                        /* possibly adapt the support range */
  CLOCK(t);                     /* start timer, print log message */
  XMSG(stderr, "writing %s ... ", isr_name(ccn->report));
  if      (simcov > CCN_NONE)   /* if a  similarity measure is given */
    r = ccn_simrepo(ccn, simcov, simps, sim, beg, end);
  else if (ccn->algo == CCN_FILTER)
    r = ccn_filter (ccn);       /* point and train filtering */
  else if (ccn->algo == CCN_REPO)
    r = ccn_cmrepo (ccn);       /* closed/maximal with a repository */
  else
    r = ccn_basic  (ccn);       /* basic algorithm (fallback) */
  if (r < 0) return E_NOMEM;    /* search for frequent item sets */
  XMSG(stderr, "[%"ISCNT_FMT" pattern(s)]",
       ISCOUT(isr_repcnt(ccn->report)));
  XMSG(stderr, " done [%.2fs].\n", SEC_SINCE(t));
  return 0;                     /* return 'ok' */
}  /* ccn_jim() */

#endif
/*----------------------------------------------------------------------
  Main Functions
----------------------------------------------------------------------*/
#ifdef CCN_MAIN

static void help (void)
{                               /* --- print add. option information */
  #ifndef QUIET
  fprintf(stderr, "\n");        /* terminate startup message */
  #ifdef CCN_OVERLAP            /* if influence map overlap support */
  printf("coconad algorithm variants (option -A#, default: -Ai)\n");
  #else                         /* if standard (MIS) support */
  printf("coconad algorithm variants (option -A#, default: -Ar/f)\n");
  #endif                        /* (different default algorithms) */
  printf("  b   basic version with linear memory consumption\n");
  printf("  f   refined version with event and train filtering\n");
  printf("  r   closed/maximal patterns filtered with a repository\n");
  printf("      (equivalent to variant 'f' for frequent patterns)\n");
  #ifdef CCN_OVERLAP            /* if influence map overlap support */
  printf("  i   interval list intersection (c/m with repository)\n");
  #else                         /* special algorithm variant */
  printf("(default is 'f' with options -Tp/s, otherwise it is 'r')\n");
  #endif                        /* default depends on target type */
  printf("information output format characters (option -v#)\n");
  #ifdef CCN_OVERLAP            /* if influence map overlap support */
  printf("item cover similarity measures (option -i#)\n");
  printf("  x   none (standard influence map overlap support)\n");
  printf("  r   Russel & Rao                S_R = s/n\n");
  printf("  k   Kulczynski                  S_K = s/q\n");
  printf("  j   Jaccard/Tanimoto            S_J = s/r\n");
  printf("  d   Dice/Sorensen/Czekanowski   S_D = 2s/(r+s)\n");
  printf("  s   Sokal & Sneath 1            S_S = s/(r+q)\n");
  printf("  m   Sokal & Michener/Hamming    S_M = (s+z)/n\n");
  printf("  f   Faith                       S_F = (s+z/2)/n\n");
  printf("  t   Rogers & Tanimoto           S_T = (s+z)/(n+q)\n");
  printf("  n   Sokal & Sneath 2            S_N = 2(s+z)/(n+s+z)\n");
  printf("  o   Sokal & Sneath 3            S_O = (s+z)/q \n");
  printf("  b   Baroni-Urbani & Buser       "
               "S_B = (sqrt(sz)+s)/(sqrt(sz)+r)\n");
  printf("  g   generic similarity measure  "
               "S   = (c0s +c1z +c2n +c3sqrt(sz))\n");
  printf("                                  "
               "    / (c4s +c5z +c6n +c7sqrt(sz))\n");
  printf("      parameters can be specified "
               "with the option -p#:#:#:#:#:#:#:#\n");
  printf("all measures are defined in the following quantities:\n");
  printf("  n   total recording time divided by window width\n");
  printf("  s   total length of intervals "
               "with all items in the item set\n");
  printf("  r   total length of intervals "
               "with at least one item in the item set\n");
  printf("  q   total length of intervals "
               "with some (but not all) items\n");
  printf("  z   total length of intervals "
               "with no items in the item set\n");
  printf("\n");
  #endif
  printf("  %%%%  a percent sign\n");
  printf("  %%i  number of items (item set size)\n");
  #ifdef CCN_OVERLAP            /* if influence map overlap support */
  printf("  if option -e is not given:\n");
  printf("  %%a  (absolute) item set support\n");
  printf("  %%w  item cover similarity as a fraction\n");
  printf("  %%W  item cover similarity as a percentage\n");
  printf("  if option -e is given:\n");
  printf("  %%w  (absolute) item set support\n");
  printf("  %%a  item cover similarity as a fraction\n");
  printf("  %%s  item cover similarity as a fraction\n");
  printf("  %%S  item cover similarity as a percentage\n");
  #else
  printf("  %%a  (absolute) item set support\n");
  #endif
  #endif                        /* print help information */
  exit(0);                      /* abort the program */
}  /* help() */

/*--------------------------------------------------------------------*/

static ITEM getbdr (char *s, char **end, RSUPP **border)
{                               /* --- get the support border */
  ITEM  i, k;                   /* loop variables */
  RSUPP *b;                     /* support border */

  assert(s && end && border);   /* check the function arguments */
  for (i = k = 0; s[i]; i++)    /* traverse the string and */
    if (s[i] == ':') k++;       /* count the number separators */
  *border = b = (RSUPP*)malloc((size_t)++k *sizeof(RSUPP));
  if (!b) return -1;            /* allocate a support border */
  for (i = 0; i < k; i++) {     /* traverse the parameters */
    b[i] = strtosupp(s, end);      /* get the next parameter and */
    if (*end == s) break;       /* check for an empty parameter */
    s = *end; if (*s++ != ':') break;
  }                             /* check for a colon */
  if (++i < k)                  /* shrink support array if possible */
    *border = (RSUPP*)realloc(b, (size_t)i *sizeof(RSUPP));
  return i;                     /* return number of support values */
}  /* getbdr() */

/*--------------------------------------------------------------------*/

static int setbdr (ISREPORT *report, ITEM zmin,
                   RSUPP **border, ITEM n)
{                               /* --- set the support border */
  RSUPP s;                      /* to traverse the support values */

  assert(report                 /* check the function arguments */
  &&    (zmin >= 0) && border && (*border || (n <= 0)));
  while (--n >= 0) {            /* traverse the support values */
    s = (*border)[n]; if (s < 0) s = -s;
    if (isr_setbdr(report, n+zmin, s) < 0) return -1;
  }                             /* set support in item set reporter */
  if (*border) { free(*border); *border = NULL; }
  return 0;                     /* return 'ok' */
}  /* setbdr() */

/*--------------------------------------------------------------------*/

#ifndef NDEBUG                  /* if debug version */
  #undef  CLEANUP               /* clean up memory and close files */
  #define CLEANUP \
  if (coconad) ccn_delete(coconad, 0); \
  if (twrite)  twr_delete(twrite, 1);  \
  if (report)  isr_delete(report, 0);  \
  if (trnset)  tns_delete(trnset, 0);  \
  if (tread)   trd_delete(tread,  1);  \
  if (ibase)   ib_delete (ibase);      \
  if (border)  free(border);
#endif

GENERROR(error, exit)           /* generic error reporting function */

/*--------------------------------------------------------------------*/

int main (int argc, char *argv[])
{                               /* --- main function */
  int     i, k = 0;             /* loop variables, counters */
  char    *s;                   /* to traverse the options */
  CCHAR   **optarg = NULL;      /* option argument */
  CCHAR   *fn_inp  = NULL;      /* name of input  file */
  CCHAR   *fn_out  = NULL;      /* name of output file */
  CCHAR   *fn_sel  = NULL;      /* name of item selection file */
  CCHAR   *fn_psp  = NULL;      /* name of pattern spectrum file */
  CCHAR   *recseps = NULL;      /* record  separators */
  CCHAR   *fldseps = NULL;      /* field   separators */
  CCHAR   *blanks  = NULL;      /* blank   characters */
  CCHAR   *comment = NULL;      /* comment characters */
  CCHAR   *hdr     = "";        /* record header  for output */
  CCHAR   *sep     = " ";       /* item separator for output */
  CCHAR   *dflt    = " (%a)";   /* default format for check */
  CCHAR   *info    = dflt;      /* format for information output */
  int     target   = 'i';       /* target pattern type (item set) */
  int     trgsub   = 'c';       /* target subtype (closed/maximal) */
  ITEM    zmin     = 2;         /* minimum size of an item set */
  ITEM    zmax     = ITEM_MAX;  /* maximum size of an item set */
  double  width    = 0.003;     /* window width/maximum distance */
  SUPP    smin     = 2;         /* minimum support of an item set */
  #ifdef CCN_OVERLAP            /* if influence map overlap support */
  int     simcov   = 'x';       /* item cover similarity measure */
  double  simps[8] = { 1, 0, 0, 0,  0, -1, 1, 0 };
  double  sim      = 10;        /* minimum similarity (in percent) */
  double  beg      = -INFINITY; /* beginning of event range */
  double  end      =  INFINITY; /* end       of event range */
  double  mix      = 0.0;       /* mixing coefficient for standard */
  #endif                        /* and influence map overlap support */
  int     algo     = 0;         /* variant of coconad algorithm */
  int     mode     = CCN_DEFAULT;  /* search mode (e.g. pruning) */
  int     minp     = TRN_ITEMPT;/* mode for train set reading */
  int     scan     = 0;         /* flag for scanable item output */
  int     bdrcnt   = 0;         /* number of support values in border */
  int     stats    = 0;         /* flag for item set statistics */
  PATSPEC *psp;                 /* collected pattern spectrum */
  ITEM    m;                    /* number of items/trains */
  size_t  n;                    /* number of events/spikes */
  #ifndef QUIET                 /* if not quiet version */
  clock_t t;                    /* timer for measurements */

  prgname = argv[0];            /* get program name for error msgs. */

  /* --- print usage message --- */
  if (argc > 1) {               /* if arguments are given */
    fprintf(stderr, "%s - %s\n", argv[0], DESCRIPTION);
    fprintf(stderr, VERSION); } /* print a startup message */
  else {                        /* if no arguments given */
    printf("usage: %s [options] infile [outfile]\n", argv[0]);
    printf("%s\n", DESCRIPTION);
    printf("%s\n", VERSION);
    #ifndef CCN_OVERLAP         /* if standard (MIS) support type */
    printf("-T#      target pattern type                      "
                    "(default: %c)\n", target);
    printf("         (i: item sets, p: partial permutations "
                    "(without repetition),\n");
    printf("          s: general sequences "
                    "(with & without repetition))\n");
    printf("-t#      target pattern subtype                   "
    #else                       /* if influence map overlap support */
    printf("-t#      target pattern type                      "
    #endif
                    "(default: %c)\n", trgsub);
    printf("         (s/f: frequent, c: closed, "
                     "m: maximal patterns,\n");
    printf("          b: closed  w.r.t. appending items,\n");
    printf("          l: maximal w.r.t. appending items)\n");
    printf("-m#      minimum number of items per item set     "
                    "(default: %"ITEM_FMT")\n", zmin);
    printf("-n#      maximum number of items per item set     "
                    "(default: no limit)\n");
    printf("-w#      width of time window/maximum distance    "
                    "(default: %g)\n", width);
    #ifndef CCN_OVERLAP         /* if standard (MIS) support type */
    printf("         (positive: maximum independent set support,\n");
    printf("          negative: non-overlapping "
                     "minimal windows support)\n");
    #endif
    printf("-s#      minimum support of an item set           "
                    "(default: %"SUPP_FMT")\n", smin);
    #ifdef CCN_OVERLAP          /* if influence map overlap support */
    printf("-M#      mixing coefficient for MIS support       "
                    "(default: %g)\n", mix);
    #endif                      /* (support mixing is possible) */
    #ifdef CCN_OVERLAP          /* if influence map overlap support */
    printf("-i#      item cover similarity measure to use     "
                    "(default: %c)\n", simcov);
    printf("-p#:#..  parameters for generic cover similarity  "
                    "(default: 1:0:0:0:0:-1:1:0)\n");
    printf("-c#      minimum cover similarity of an item set  "
                    "(default: %g%%)\n", sim);
    printf("-a#      beginning of event/spike range           "
                    "(default: -oo)\n");
    printf("-z#      end       of event/spike range           "
                    "(default: +oo)\n");
    printf("-e       exchange roles of support and similarity\n");
    printf("         (for reporting and pattern spectrum)\n");
    #endif
    printf("-A#      variant of the algorithm to use          "
    #ifndef CCN_OVERLAP         /* if standard (MIS) support type */
                    "(default: r/f)\n");
    printf("         (default is r with option -Ti "
                            "and f with options -Tp/-Ts)\n");
    #else                       /* if influence map overlap support */
                    "(default: i)\n");
    #endif                      /* (different algorithm defaults) */
    printf("-0       list/array for the repository            "
                    "(default: tree chain)\n");
    #ifndef CCN_OVERLAP         /* if standard (MIS) support type */
    printf("-1       single prefix tree for the repository    "
                    "(default: tree chain)\n");
    printf("         (only for partial permutations and sequences)\n");
    printf("-2       prefix tree chain for the repository     "
                    "(default: tree chain)\n");
    printf("-p       do not sort trains by their length       "
                    "(default: sort)\n");
    #endif                      /* sorting by train length may help */
    printf("-x       do not prune with perfect extensions     "
                    "(default: prune)\n");
    printf("-l       one train per record (item + events)     "
                    "(default: id/time pairs)\n");
    printf("-y       no item (with -l) or item after time     "
                    "(default: item first)\n");
    printf("-F#:#..  support border for filtering item sets   "
                    "(default: none)\n");
    printf("         (list of minimum support values, "
                    "one per item set size,\n");
    printf("         starting at the minimum size, "
                    "as given with option -m#)\n");
    #ifdef CCN_OVERLAP
    printf("         (minimum similarity values "
                    "if item cover similarity is used)\n");
    #endif
    printf("-R#      read an item selection from a file       "
                    "(default: use all items)\n");
    printf("-P#      write a pattern spectrum to a file\n");
    printf("-Z       print item set statistics "
                    "(number of item sets per size)\n");
    printf("-h#      record header  for output                "
                    "(default: \"%s\")\n", hdr);
    printf("-k#      item separator for output                "
                    "(default: \"%s\")\n", sep);
    printf("-v#      output format for item set information   "
                    "(default: \"%s\")\n", info);
    printf("-r#      record/transaction separators            "
                    "(default: \"\\n\")\n");
    printf("-f#      field /item        separators            "
                    "(default: \" \\t,\")\n");
    printf("-b#      blank   characters                       "
                    "(default: \" \\t\\r\")\n");
    printf("-C#      comment characters                       "
                    "(default: \"#\")\n");
    printf("-!       print additional option information\n");
    printf("infile   file to read trains from                 "
                    "[required]\n");
    printf("outfile  file to write pattern spectrum to        "
                    "[optional]\n");
    return 0;                   /* print a usage message */
  }                             /* and abort the program */
  #endif  /* #ifndef QUIET */

  /* --- evaluate arguments --- */
  for (i = 1; i < argc; i++) {  /* traverse arguments */
    s = argv[i];                /* get option argument */
    if (optarg) { *optarg = s; optarg = NULL; continue; }
    if ((*s == '-') && *++s) {  /* -- if argument is an option */
      while (*s) {              /* traverse options */
        switch (*s++) {         /* evaluate switches */
          case '!': help();                          break;
          #ifndef CCN_OVERLAP   /* if standard (MIS) support */
          case 'T': target = (*s) ? *s++ : 's';      break;
          #endif
          case 't': trgsub = (*s) ? *s++ : 's';      break;
          #ifdef CCN_OVERLAP    /* if influence map overlap support */
          case 'M': mix    =       strtod(s, &s);    break;
          #endif                /* sequence mining is possible */
          case 'm': zmin   = (ITEM)strtol(s, &s, 0); break;
          case 'n': zmax   = (ITEM)strtol(s, &s, 0); break;
          case 'w': width  =       strtod(s, &s);    break;
          case 's': smin   =    strtosupp(s, &s);    break;
          #ifdef CCN_OVERLAP    /* if influence map overlap support */
          case 'i': simcov = (*s) ? *s++ : 'x';      break;
          case 'p': getdblvec(s, &s, 8, simps);      break;
          case 'c': sim    =       strtod(s, &s);    break;
          case 'a': beg    =       strtod(s, &s);    break;
          case 'z': end    =       strtod(s, &s);    break;
          case 'e': mode  |=  CCN_SIMSUPP;           break;
          #endif                /* cover similarity is possible */
          case 'A': algo   = (*s) ? *s++ : 0;        break;
          case '0': mode  |=  CCN_REPOLIST;          break;
          #ifndef CCN_OVERLAP   /* if standard (MIS) support */
          case '1': mode  |=  CCN_REPOTREE;          break;
          case '2': mode  |=  CCN_REPOCHAIN;         break;
          #endif                /* (single prefix tree repository) */
          case 'x': mode  &= ~CCN_PERFECT;           break;
          #ifndef CCN_OVERLAP   /* if standard (MIS) support type */
          case 'p': mode  &= ~CCN_REORDER;           break;
          #endif
          case 'l': minp  &= ~TRN_ITEMPT;            break;
          case 'y': minp  |=  TRN_PURETRN;           break;
          case 'F': bdrcnt = getbdr(s, &s, &border); break;
          case 'R': optarg = &fn_sel;                break;
          case 'P': optarg = &fn_psp;                break;
          case 'Z': stats  = 1;                      break;
          case 'h': optarg = &hdr;                   break;
          case 'k': optarg = &sep;                   break;
          case 'v': optarg = &info;                  break;
          case 'r': optarg = &recseps;               break;
          case 'f': optarg = &fldseps;               break;
          case 'b': optarg = &blanks;                break;
          case 'C': optarg = &comment;               break;
          default : error(E_OPTION, *--s);           break;
        }                       /* set option variables */
        if (optarg && *s) { *optarg = s; optarg = NULL; break; }
      } }                       /* get option argument */
    else {                      /* -- if argument is no option */
      switch (k++) {            /* evaluate non-options */
        case  0: fn_inp = s;      break;
        case  1: fn_out = s;      break;
        default: error(E_ARGCNT); break;
      }                         /* note filenames */
    }
  }
  if (optarg)      error(E_OPTARG);     /* check option arguments */
  if (k < 1)       error(E_ARGCNT);     /* and number of arguments */
  #ifdef CCN_OVERLAP            /* if influence map overlap support */
  if (width  <= 0) error(E_WIDTH,   width);
  #endif                        /* window width may not be negative */
  if (smin   == 0) error(E_SUPPORT, smin);
  if (smin   <  0) smin = -smin;/* ensure positive minimum support */
  if (zmin   <  1) error(E_SIZE, zmin); /* check the size limits */
  if (zmax   <  1) error(E_SIZE, zmax);
  if (bdrcnt <  0) error(E_NOMEM);
  switch (target) {             /* check and translate target type */
    case 'i': target = CCN_ITEMSET;          break;
    case 'p': target = CCN_PERMUT;           break;
    case 's': target = CCN_SEQUENCE;         break;
    default : error(E_TARGET, (char)target); break;
  }
  switch (trgsub) {             /* check and translate target subtype */
    case 's': target |= CCN_FREQUENT;           break;
    case 'f': target |= CCN_FREQUENT;           break;
    case 'c': target |= CCN_CLOSED;             break;
    case 'm': target |= CCN_MAXIMAL;            break;
    case 'b': target |= CCN_APPEND|CCN_CLOSED;  break;
    case 'l': target |= CCN_APPEND|CCN_MAXIMAL; break;
    default : error(E_TARGET, (char)trgsub);    break;
  }                             /* (get target type code) */
  #ifdef CCN_OVERLAP            /* if influence map overlap support */
  switch (simcov) {             /* check and translate sim. measure */
    case 'x': simcov = CCN_NONE;              break;
    case 'r': simcov = CCN_RUSSEL_RAO;        break;
    case 'k': simcov = CCN_KULCZYNSKI;        break;
    case 'j': simcov = CCN_JACCARD;           break;
    case 'd': simcov = CCN_DICE;              break;
    case 's': simcov = CCN_SOKAL_SNEATH_1;    break;
    case 'm': simcov = CCN_SOKAL_MICHENER;    break;
    case 'f': simcov = CCN_FAITH;             break;
    case 't': simcov = CCN_ROGERS_TANIMOTO;   break;
    case 'n': simcov = CCN_SOKAL_SNEATH_2;    break;
    case 'o': simcov = CCN_SOKAL_SNEATH_3;    break;
    case 'b': simcov = CCN_BARONI_BUSER;      break;
    case 'g': simcov = CCN_GENERIC;           break;
    default : error(E_SIMILTY, (char)simcov); break;
  }                             /* (get cover similarity measure) */
  #endif
  #ifndef CCN_OVERLAP           /* if integer support type */
  if (algo == 0)                /* choose default algorithm */
    algo = (target & (CCN_PERMUT|CCN_SEQUENCE)) ? 'f' : 'r';
  #else                         /* if influence map overlap support */
  if (algo == 0) algo = 'i';    /* choose default algorithm */
  #endif
  switch (algo) {               /* check and translate alg. variant */
    case 'b': algo = CCN_BASIC;               break;
    case 'f': algo = CCN_FILTER;              break;
    case 'r': algo = CCN_REPO;                break;
    #ifdef CCN_OVERLAP          /* if influence map overlap support */
    case 'i': algo = CCN_ISECT;               break;
    #endif                      /* additional coconad variant(s) */
    default : error(E_VARIANT, (char)algo);   break;
  }                             /* (get CoCoNAD algorithm code) */
  MSG(stderr, "\n");            /* terminate the startup message */
  mode |= CCN_NOCLEAN|CCN_VERBOSE;

  /* --- read item selection --- */
  ibase = ib_create(0, 0);      /* create an item base */
  if (!ibase) error(E_NOMEM);   /* to manage the items */
  tread = trd_create();         /* create a transaction reader */
  if (!tread) error(E_NOMEM);   /* and configure the characters */
  trd_allchs(tread, recseps, fldseps, blanks, "", comment);
  if (fn_sel) {                 /* if item appearances are given */
    CLOCK(t);                   /* start timer, open input file */
    if (trd_open(tread, NULL, fn_sel) != 0)
      error(E_FOPEN, trd_name(tread));
    MSG(stderr, "reading %s ... ", trd_name(tread));
    m = ib_readsel(ibase,tread);/* read the given item selection */
    if (m < 0) error((int)-m, ib_errmsg(ibase, NULL, 0));
    trd_close(tread);           /* close the input file */
    MSG(stderr, "[%"ITEM_FMT" item(s)]", m);
    MSG(stderr, " done [%.2fs].\n", SEC_SINCE(t));
  }                             /* print a log message */

  /* --- read train set --- */
  trnset = tns_create(ibase);   /* create a train set */
  if (!trnset) error(E_NOMEM);  /* to store the trains */
  CLOCK(t);                     /* start timer, open input file */
  if (trd_open(tread, NULL, fn_inp) != 0)
    error(E_FOPEN, trd_name(tread));
  MSG(stderr, "reading %s ... ", trd_name(tread));
  k = tns_read(trnset, tread, minp);
  if (k < 0) error(-k, tns_errmsg(trnset, NULL, 0));
  trd_delete(tread, 1);         /* read the trains database, */
  tread = NULL;                 /* then delete the table reader */
  m = tns_cnt(trnset);          /* get the number of trains */
  n = tns_extent(trnset);       /* and the number of events */
  MSG(stderr, "[%"ITEM_FMT" train(s), %"SIZE_FMT" event(s)]", m, n);
  MSG(stderr, " done [%.2fs].", SEC_SINCE(t));
  if ((m <= 0) || (n <= 0))     /* check for at least one item */
    error(E_NOITEMS);           /* and at least one event/spike */
  MSG(stderr, "\n");            /* terminate the log message */

  /* --- find frequent item sets --- */
  coconad = ccn_create(target, width, smin,
  #ifdef CCN_OVERLAP            /* if influence map overlap support */
                       mix,     /* add. support mixing parameter */
  #endif
                       zmin, zmax, algo, mode);
  if (!coconad) error(E_NOMEM); /* create a coconad miner */
  k = ccn_data(coconad, trnset, 0);
  if (k) error(k);              /* prepare data for CoCoNAD */
  #ifndef CCN_OVERLAP           /* if standard (MIS) support */
  if (target & CCN_SEQUENCE)    /* if sequences (with poss. reps.) */
    m = tns_maxseq(trnset, fabs(width)) +1;
  #endif                        /* find max. possible sequence len. */
  report = isr_createx(ibase, m);
  if (!report) error(E_NOMEM);  /* create an item set reporter */
  k = ccn_report(coconad, report);
  if (k) error(k);              /* prepare reporter for coconad */
  if (setbdr(report, zmin, &border, bdrcnt) != 0)
    error(E_NOMEM);             /* set the support border (if any) */
  if (fn_psp && (isr_addpsp(report, NULL) < 0))
    error(E_NOMEM);             /* set a pattern spectrum if req. */
  if (isr_setfmt(report, scan, hdr, sep, NULL, info) != 0)
    error(E_NOMEM);             /* set the output format strings */
  k = isr_open(report, NULL, fn_out);
  if (k) error(k, isr_name(report));
  if (isr_setup(report) < 0)    /* open the item set file and */
    error(E_NOMEM);             /* set up the item set reporter */
  #ifdef CCN_OVERLAP            /* if influence map overlap support */
  if (simcov > 0)               /* if a  similarity measure is given */
    k = ccn_jim (coconad, simcov, simps, sim, beg, end);
  else                          /* if no similarity measure is given */
    k = ccn_mine(coconad);      /* find frequent item patterns */
  #else                         /* or if standard (MIS) support */
  k = ccn_mine(coconad);        /* find frequent item patterns */
  #endif
  if (k) error(k);              /* check for a mining error */
  if (stats)                    /* print item set statistics */
    isr_prstats(report, stdout, 0);
  if (isr_close(report) != 0)   /* close item set output file */
    error(E_FWRITE, isr_name(report));

  /* --- write pattern spectrum --- */
  if (fn_psp) {                 /* if to write a pattern spectrum */
    CLOCK(t);                   /* start timer, create table write */
    psp    = isr_getpsp(report);/* get the pattern spectrum */
    twrite = twr_create();      /* create a table writer and */
    if (!twrite) error(E_NOMEM);/* open the output file */
    if (twr_open(twrite, NULL, fn_psp) != 0)
      error(E_FOPEN,  twr_name(twrite));
    MSG(stderr, "writing %s ... ", twr_name(twrite));
    if (psp_report(psp, twrite, 1.0) != 0)
      error(E_FWRITE, twr_name(twrite));
    twr_delete(twrite, 1);      /* write the pattern spectrum */
    twrite = NULL;              /* and delete the table writer */
    MSG(stderr, "[%"SIZE_FMT" signature(s)]", psp_sigcnt(psp));
    MSG(stderr, " done [%.2fs].\n", SEC_SINCE(t));
  }                             /* write a log message */

  /* --- clean up --- */
  CLEANUP;                      /* clean up memory and close files */
  SHOWMEM;                      /* show (final) memory usage */
  return 0;                     /* return 'ok' */
}  /* main() */

#endif  /* #ifdef CCN_MAIN */
